import pytest
from numpy.testing import assert_almost_equal, assert_equal, assert_allclose
import numpy as np
import sys
import os
import MDAnalysis as mda
from MDAnalysis.coordinates.H5MD import HAS_H5PY

if HAS_H5PY:
    import h5py
    from MDAnalysis.coordinates.H5MD import H5MDReader
from MDAnalysis.exceptions import NoDataError
from MDAnalysisTests.datafiles import (
    H5MD_xvf,
    TPR_xvf,
    TRR_xvf,
    COORDINATES_TOPOLOGY,
    COORDINATES_H5MD,
    H5MD_energy,
    H5MD_malformed,
)
from MDAnalysisTests.coordinates.base import (
    MultiframeReaderTest,
    BaseReference,
    BaseWriterTest,
)


@pytest.mark.skipif(not HAS_H5PY, reason="h5py not installed")
class H5MDReference(BaseReference):
    """Reference synthetic trajectory that was
    copied from test_xdr.TRRReference"""

    def __init__(self):
        super(H5MDReference, self).__init__()
        self.trajectory = COORDINATES_H5MD
        self.topology = COORDINATES_TOPOLOGY
        self.reader = mda.coordinates.H5MD.H5MDReader
        self.writer = mda.coordinates.H5MD.H5MDWriter
        self.ext = "h5md"
        self.prec = 3
        self.changing_dimensions = True

        self.first_frame.velocities = self.first_frame.positions / 10
        self.first_frame.forces = self.first_frame.positions / 100

        self.second_frame.velocities = self.second_frame.positions / 10
        self.second_frame.forces = self.second_frame.positions / 100

        self.last_frame.velocities = self.last_frame.positions / 10
        self.last_frame.forces = self.last_frame.positions / 100

        self.jump_to_frame.velocities = self.jump_to_frame.positions / 10
        self.jump_to_frame.forces = self.jump_to_frame.positions / 100

    def iter_ts(self, i):
        ts = self.first_frame.copy()
        ts.positions = 2**i * self.first_frame.positions
        ts.velocities = ts.positions / 10
        ts.forces = ts.positions / 100
        ts.time = i
        ts.frame = i
        return ts


@pytest.mark.skipif(not HAS_H5PY, reason="h5py not installed")
class TestH5MDReaderBaseAPI(MultiframeReaderTest):
    """Tests H5MDReader with with synthetic trajectory."""

    @staticmethod
    @pytest.fixture()
    def ref():
        return H5MDReference()

    def test_get_writer_1(self, ref, reader, tmpdir):
        with tmpdir.as_cwd():
            outfile = "test-writer." + ref.ext
            with reader.Writer(outfile) as W:
                assert_equal(isinstance(W, ref.writer), True)
                assert_equal(W.n_atoms, reader.n_atoms)

    def test_get_writer_2(self, ref, reader, tmpdir):
        with tmpdir.as_cwd():
            outfile = "test-writer." + ref.ext
            with reader.Writer(outfile, n_atoms=100) as W:
                assert_equal(isinstance(W, ref.writer), True)
                assert_equal(W.n_atoms, 100)

    def test_copying(self, ref, reader):
        # Issue #3664 - test not done in test_copying due to dependencies
        original = mda.coordinates.H5MD.H5MDReader(
            ref.trajectory,
            convert_units=False,
            dt=2,
            time_offset=10,
            foo="bar",
        )
        copy = original.copy()

        assert original.format not in ("MEMORY", "CHAIN")
        assert original.convert_units is False
        assert copy.convert_units is False
        assert original._ts_kwargs["time_offset"] == 10
        assert copy._ts_kwargs["time_offset"] == 10
        assert original._ts_kwargs["dt"] == 2
        assert copy._ts_kwargs["dt"] == 2

        assert original.ts.data["time_offset"] == 10
        assert copy.ts.data["time_offset"] == 10

        assert original.ts.data["dt"] == 2
        assert copy.ts.data["dt"] == 2

        assert copy._kwargs["foo"] == "bar"

        # check coordinates
        assert original.ts.frame == copy.ts.frame
        assert_allclose(original.ts.positions, copy.ts.positions)

        original.next()
        copy.next()

        assert original.ts.frame == copy.ts.frame
        assert_allclose(original.ts.positions, copy.ts.positions)


@pytest.mark.skipif(not HAS_H5PY, reason="h5py not installed")
class TestH5MDWriterBaseAPI(BaseWriterTest):
    """Tests H5MDWriter base API with synthetic trajectory"""

    @staticmethod
    @pytest.fixture()
    def ref():
        return H5MDReference()

    def test_write_trajectory_atomgroup(self, ref, reader, universe, tmpdir):
        outfile = "write-atoms-test." + ref.ext
        with tmpdir.as_cwd():
            with ref.writer(
                outfile, universe.atoms.n_atoms, velocities=True, forces=True
            ) as w:
                for ts in universe.trajectory:
                    w.write(universe.atoms)
            self._check_copy(outfile, ref, reader)

    @pytest.mark.xfail(
        (os.name == "nt" and sys.maxsize <= 2**32),
        reason="occasional fail on 32-bit windows",
    )
    def test_write_trajectory_universe(self, ref, reader, universe, tmpdir):
        outfile = "write-uni-test." + ref.ext
        with tmpdir.as_cwd():
            with ref.writer(
                outfile, universe.atoms.n_atoms, velocities=True, forces=True
            ) as w:
                for ts in universe.trajectory:
                    w.write(universe)
            self._check_copy(outfile, ref, reader)


@pytest.mark.skipif(not HAS_H5PY, reason="h5py not installed")
class TestH5MDReaderWithRealTrajectory(object):

    prec = 3
    ext = "h5md"

    @pytest.fixture(scope="class")
    def universe(self):
        return mda.Universe(TPR_xvf, H5MD_xvf)

    @pytest.fixture()
    def h5md_file(self):
        return h5py.File(H5MD_xvf, "r")

    @pytest.fixture()
    def outfile(self, tmpdir):
        return str(tmpdir.join("h5md-reader-test." + self.ext))

    def test_n_frames(self, universe):
        assert len(universe.trajectory) == 3

    def test_positions(self, universe):
        universe.trajectory[0]
        assert_almost_equal(
            universe.atoms.positions[0],
            [32.309906, 13.77798, 14.372463],
            decimal=self.prec,
        )
        assert_almost_equal(
            universe.atoms.positions[42],
            [28.116928, 19.405945, 19.647358],
            decimal=self.prec,
        )
        assert_almost_equal(
            universe.atoms.positions[10000],
            [44.117805, 50.442093, 23.299038],
            decimal=self.prec,
        )

        universe.trajectory[1]
        assert_almost_equal(
            universe.atoms.positions[0],
            [30.891968, 13.678971, 13.6000595],
            decimal=self.prec,
        )
        assert_almost_equal(
            universe.atoms.positions[42],
            [27.163246, 19.846561, 19.3582],
            decimal=self.prec,
        )
        assert_almost_equal(
            universe.atoms.positions[10000],
            [45.869278, 5.0342298, 25.460655],
            decimal=self.prec,
        )
        universe.trajectory[2]
        assert_almost_equal(
            universe.atoms.positions[0],
            [31.276512, 13.89617, 15.015897],
            decimal=self.prec,
        )
        assert_almost_equal(
            universe.atoms.positions[42],
            [28.567991, 20.56532, 19.40814],
            decimal=self.prec,
        )
        assert_almost_equal(
            universe.atoms.positions[10000],
            [39.713223, 6.127234, 18.284992],
            decimal=self.prec,
        )

    def test_h5md_velocities(self, universe):
        universe.trajectory[0]
        assert_almost_equal(
            universe.atoms.velocities[0],
            [-2.697732, 0.613568, 0.14334752],
            decimal=self.prec,
        )
        universe.trajectory[1]
        assert_almost_equal(
            universe.atoms.velocities[42],
            [-6.8698354, 7.834235, -8.114698],
            decimal=self.prec,
        )
        universe.trajectory[2]
        assert_almost_equal(
            universe.atoms.velocities[10000],
            [9.799492, 5.631466, 6.852126],
            decimal=self.prec,
        )

    def test_h5md_forces(self, universe):
        universe.trajectory[0]
        assert_almost_equal(
            universe.atoms.forces[0],
            [20.071287, -155.2285, -96.72112],
            decimal=self.prec,
        )
        universe.trajectory[1]
        assert_almost_equal(
            universe.atoms.forces[42],
            [-4.1959066, -31.31548, 22.663044],
            decimal=self.prec,
        )
        universe.trajectory[2]
        assert_almost_equal(
            universe.atoms.forces[10000],
            [-41.43743, 83.35207, 62.94751],
            decimal=self.prec,
        )

    def test_h5md_dimensions(self, universe):
        universe.trajectory[0]
        assert_almost_equal(
            universe.trajectory.ts.dimensions,
            [52.763, 52.763, 52.763, 90.0, 90.0, 90.0],
            decimal=self.prec,
        )
        universe.trajectory[1]
        assert_almost_equal(
            universe.trajectory.ts.dimensions,
            [52.807877, 52.807877, 52.807877, 90.0, 90.0, 90.0],
            decimal=self.prec,
        )
        universe.trajectory[2]
        assert_almost_equal(
            universe.trajectory.ts.dimensions,
            [52.839806, 52.839806, 52.839806, 90.0, 90.0, 90.0],
            decimal=self.prec,
        )

    def test_h5md_data_step(self, universe):
        for ts, step in zip(universe.trajectory, (0, 25000, 50000)):
            assert_equal(ts.data["step"], step)

    def test_rewind(self, universe):
        universe.trajectory[1]
        universe.trajectory.rewind()
        assert universe.trajectory.ts.frame == 0

    def test_next(self, universe):
        universe.trajectory.rewind()
        universe.trajectory.next()
        assert universe.trajectory.ts.frame == 1

    def test_jump_last_frame(self, universe):
        universe.trajectory[-1]
        assert universe.trajectory.ts.frame == 2

    @pytest.mark.parametrize("start, stop, step", ((0, 2, 1), (1, 2, 1)))
    def test_slice(self, universe, start, stop, step):
        frames = [
            universe.trajectory.ts.frame
            for ts in universe.trajectory[start:stop:step]
        ]
        assert_equal(frames, np.arange(start, stop, step))

    @pytest.mark.parametrize("array_like", [list, np.array])
    def test_array_like(self, universe, array_like):
        array = array_like([0, 2])
        frames = [
            universe.trajectory.ts.frame for ts in universe.trajectory[array]
        ]
        assert_equal(frames, array)

    def test_list_indices(self, universe):
        indices = [0, 1, 2, 1, 2, 2, 0]
        frames = [
            universe.trajectory.ts.frame for ts in universe.trajectory[indices]
        ]
        assert_equal(frames, indices)

    @pytest.mark.parametrize(
        "group, attr",
        (
            ("position", "positions"),
            ("velocity", "velocities"),
            ("force", "forces"),
        ),
    )
    def test_no_group(self, h5md_file, outfile, attr, group):
        with h5md_file as f:
            with h5py.File(outfile, "w") as g:
                f.copy(source="particles", dest=g)
                f.copy(source="h5md", dest=g)
                del g[f"particles/trajectory/{group}"]
        u = mda.Universe(TPR_xvf, outfile)
        with pytest.raises(NoDataError, match="This Timestep has no"):
            getattr(u.trajectory.ts, attr)

    @pytest.mark.parametrize(
        "dset",
        ("position/value", "position/time", "velocity/value", "force/value"),
    )
    def test_unknown_unit(self, h5md_file, outfile, dset):
        with h5md_file as f:
            with h5py.File(outfile, "w") as g:
                f.copy(source="particles", dest=g)
                f.copy(source="h5md", dest=g)
                g["particles" "/trajectory" f"/{dset}"].attrs[
                    "unit"
                ] = "random string"
        with pytest.raises(
            RuntimeError, match=" is not recognized by H5MDReader."
        ):
            u = mda.Universe(TPR_xvf, outfile)

    def test_length_unit_from_box(self, h5md_file, universe, outfile):
        with h5md_file as f:
            with h5py.File(outfile, "w") as g:
                f.copy(source="particles", dest=g)
                f.copy(source="h5md", dest=g)
                del g["particles/trajectory/position"]
        ref_u = universe
        uw = mda.Universe(TPR_xvf, outfile)
        assert_equal(
            ref_u.trajectory.units["length"], uw.trajectory.units["length"]
        )
        for ref_ts, new_ts in zip(ref_u.trajectory, uw.trajectory):
            assert_equal(ref_ts.dimensions, new_ts.dimensions)
            assert_equal(
                ref_ts.triclinic_dimensions, new_ts.triclinic_dimensions
            )

    @pytest.mark.parametrize("group", ("position", "velocity", "force"))
    def test_changing_n_atoms(self, h5md_file, outfile, group):
        with h5md_file as f:
            with h5py.File(outfile, "w") as g:
                f.copy(source="particles", dest=g)
                f.copy(source="h5md", dest=g)
                g[f"particles/trajectory/{group}/value"].resize((3, 10000, 3))
        with pytest.raises(
            ValueError, match=" of either the postion, velocity, or force"
        ):
            u = mda.Universe(TPR_xvf, outfile)

    def test_2D_box(self, h5md_file, outfile):
        with h5md_file as f:
            with h5py.File(outfile, "w") as g:
                f.copy(source="particles", dest=g)
                f.copy(source="h5md", dest=g)
                new_box = np.ones(shape=(3, 2, 2))
                g["particles/trajectory/box"].attrs["dimension"] = 2
                del g["particles/trajectory/box/edges/value"]
                g["particles/trajectory" "/box/edges"].create_dataset(
                    "value", data=new_box
                )
        with pytest.raises(
            ValueError, match="MDAnalysis only supports 3-dimensional"
        ):
            u = mda.Universe(TPR_xvf, outfile)

    def test_box_vector(self, h5md_file, outfile):
        with h5md_file as f:
            with h5py.File(outfile, "w") as g:
                f.copy(source="particles", dest=g)
                f.copy(source="h5md", dest=g)
                vector = [1, 2, 3]
                del g["particles/trajectory/box/edges"]
                g["particles/trajectory/box/edges/value"] = [
                    vector,
                    vector,
                    vector,
                ]
        u = mda.Universe(TPR_xvf, outfile)
        # values in vector are conveted from nm -> Angstrom
        assert_equal(u.trajectory.ts.dimensions, [10, 20, 30, 90, 90, 90])

    def test_no_box(self, h5md_file, outfile):
        with h5md_file as f:
            with h5py.File(outfile, "w") as g:
                f.copy(source="particles", dest=g)
                f.copy(source="h5md", dest=g)
                del g["particles/trajectory/box/edges"]
        u = mda.Universe(TPR_xvf, outfile)
        assert_equal(u.trajectory.ts.dimensions, None)

    def test_no_groups(self, h5md_file, outfile):
        with h5md_file as f:
            with h5py.File(outfile, "w") as g:
                f.copy(source="particles", dest=g)
                f.copy(source="h5md", dest=g)
                del g["particles/trajectory/position"]
                del g["particles/trajectory/velocity"]
                del g["particles/trajectory/force"]
        with pytest.raises(
            NoDataError, match="Provide at least a position, velocity"
        ):
            u = mda.Universe(TPR_xvf, outfile)

    def test_no_convert_units(self, h5md_file, outfile):
        with h5md_file as f:
            with h5py.File(outfile, "w") as g:
                f.copy(source="particles", dest=g)
                f.copy(source="h5md", dest=g)
                groups = ["position", "velocity", "force"]
                for name in groups:
                    del g["particles/trajectory"][name]["value"].attrs["unit"]
                del g["particles/trajectory/position/time"].attrs["unit"]
        u = mda.Universe(TPR_xvf, outfile, convert_units=False)
        for unit in u.trajectory.units:
            assert_equal(u.trajectory.units[unit], None)

    def test_no_units_but_convert_units_true_error(self, h5md_file, outfile):
        with h5md_file as f:
            with h5py.File(outfile, "w") as g:
                f.copy(source="particles", dest=g)
                f.copy(source="h5md", dest=g)
                groups = ["position", "velocity", "force"]
                for name in groups:
                    del g["particles/trajectory"][name]["value"].attrs["unit"]
                del g["particles/trajectory/position/time"].attrs["unit"]
                del g["particles/trajectory/box/edges/value"].attrs["unit"]
        with pytest.raises(
            ValueError, match="H5MD file must have readable units if"
        ):
            u = mda.Universe(TPR_xvf, outfile, convert_units=True)

    @pytest.mark.xfail(reason="Issue #2884")
    def test_open_filestream(self, h5md_file, universe):
        with h5md_file as f:
            u = mda.Universe(TPR_xvf, h5md_file)
            for ts1, ts2 in zip(universe.trajectory, u.trajectory):
                assert_equal(ts1.positions, ts2.positions)
                assert_equal(ts1.velocities, ts2.velocities)
                assert_equal(ts1.forces, ts2.forces)

    def test_wrong_driver(self):
        with pytest.raises(
            ValueError, match="If MPI communicator object is used to open"
        ):
            u = mda.Universe(
                TPR_xvf,
                H5MD_xvf,
                driver="wrong_driver",
                comm="mock MPI.COMM_WORLD",
            )

    def test_open_with_driver(self):
        u = mda.Universe(TPR_xvf, H5MD_xvf, driver="core")
        assert_equal(u.trajectory._file.driver, "core")

    @pytest.mark.parametrize(
        "group1, group2",
        (
            ("velocity", "force"),
            ("position", "force"),
            ("position", "velocity"),
        ),
    )
    def test_parse_n_atoms(self, h5md_file, outfile, group1, group2):
        with h5md_file as f:
            with h5py.File(outfile, "w") as g:
                f.copy(source="particles", dest=g)
                f.copy(source="h5md", dest=g)
                traj_group = g["particles/trajectory"]
                del traj_group[group1]
                del traj_group[group2]
                for dset in (
                    "position/value",
                    "velocity/value",
                    "force/value",
                ):
                    try:
                        n_atoms_in_dset = traj_group[dset].shape[1]
                        break
                    except KeyError:
                        continue

        u = mda.Universe(outfile, to_guess=())
        assert_equal(u.atoms.n_atoms, n_atoms_in_dset)

    def test_parse_n_atoms_error(self, h5md_file, outfile):
        with h5md_file as f:
            with h5py.File(outfile, "w") as g:
                f.copy(source="particles", dest=g)
                f.copy(source="h5md", dest=g)
                traj_group = g["particles/trajectory"]
                del traj_group["position"]
                del traj_group["velocity"]
                del traj_group["force"]

        errmsg = "Could not construct minimal topology"
        with pytest.raises(ValueError, match=errmsg):
            u = mda.Universe(outfile)


@pytest.mark.skipif(not HAS_H5PY, reason="h5py not installed")
class TestH5MDWriterWithRealTrajectory(object):

    prec = 3

    @pytest.fixture()
    def universe(self):
        return mda.Universe(TPR_xvf, H5MD_xvf)

    @pytest.fixture()
    def universe_no_units(self):
        u = mda.Universe(TPR_xvf, H5MD_xvf, convert_units=False)
        u.trajectory.units["time"] = None
        u.trajectory.units["length"] = None
        u.trajectory.units["velocity"] = None
        u.trajectory.units["force"] = None
        return u

    @pytest.fixture()
    def Writer(self):
        return mda.coordinates.H5MD.H5MDWriter

    @pytest.fixture()
    def outfile(self, tmpdir):
        return str(tmpdir) + "h5md-writer-test.h5md"

    @pytest.fixture()
    def outtop(self, tmpdir):
        return str(tmpdir) + "h5md-writer-top.pdb"

    @pytest.mark.parametrize(
        "scalar, error, match",
        (
            (0, ValueError, "H5MDWriter: no atoms in output trajectory"),
            (0.5, IOError, "H5MDWriter: Timestep does not have"),
        ),
    )
    def test_n_atoms_errors(
        self, universe, Writer, outfile, scalar, error, match
    ):
        n_atoms = universe.atoms.n_atoms * scalar
        with pytest.raises(error, match=match):
            with Writer(outfile, n_atoms) as W:
                W.write(universe)

    def test_chunk_error(self, universe, Writer, outfile):
        n_atoms = universe.atoms.n_atoms
        err = "H5MDWriter must know how many frames will be "
        with pytest.raises(ValueError, match=err):
            with Writer(outfile, n_atoms, chunks=False) as W:
                for ts in universe.trajectory:
                    W.write(universe)

    @pytest.mark.parametrize("dimensions", (None, 0))
    def test_no_dimensions(self, universe, Writer, outfile, dimensions):
        with Writer(outfile, universe.atoms.n_atoms) as W:
            for ts in universe.trajectory:
                ts.dimensions = dimensions
                W.write(universe)

        uw = mda.Universe(TPR_xvf, outfile)
        box = uw.trajectory._particle_group["box"]
        assert "edges" not in box
        assert_equal(3 * ["none"], box.attrs["boundary"])
        assert_equal(3, box.attrs["dimension"])

    def test_step_not_monotonic(self, universe, Writer, outfile):
        with pytest.raises(
            ValueError, match="The H5MD standard dictates that the step "
        ):
            with Writer(outfile, universe.atoms.n_atoms) as W:
                for ts in universe.trajectory[[0, 1, 2, 1]]:
                    W.write(universe)

        with pytest.raises(
            ValueError, match="The H5MD standard dictates that the step "
        ):
            with Writer(outfile, universe.atoms.n_atoms) as W:
                for ts in universe.trajectory:
                    if ts.frame == 2:
                        ts.data["step"] = 0
                    W.write(universe)

    def test_step_from_frame(self, universe, Writer, outfile):
        with Writer(outfile, universe.atoms.n_atoms) as W:
            for ts in universe.trajectory:
                del ts.data["step"]
                W.write(universe)

        uw = mda.Universe(TPR_xvf, outfile)
        steps = [ts.data["step"] for ts in uw.trajectory]
        frames = [ts.frame for ts in universe.trajectory]
        for step, frame in zip(steps, frames):
            assert_equal(step, frame)

    def test_has_property(self, universe, Writer, outfile):
        with Writer(outfile, universe.atoms.n_atoms) as W:
            W.write(universe)
            # make sure property is pulled from _has dict
            assert W.has_positions == W._has["position"]
            assert W.has_velocities == W._has["velocity"]
            assert W.has_forces == W._has["force"]
            # make sure the values are correct
            assert W.has_positions is True
            assert W.has_velocities is True
            assert W.has_forces is True

    @pytest.mark.parametrize(
        "pos, vel, force",
        (
            (True, False, False),
            (True, True, False),
            (True, False, True),
            (True, True, True),
            (False, True, True),
            (False, False, True),
            (False, False, False),
        ),
    )
    def test_write_trajectory(
        self, universe, Writer, outfile, pos, vel, force
    ):
        try:
            with Writer(
                outfile,
                universe.atoms.n_atoms,
                positions=pos,
                velocities=vel,
                forces=force,
                author="My Name",
                author_email="my_email@asu.edu",
            ) as W:
                for ts in universe.trajectory:
                    W.write(universe)

            uw = mda.Universe(TPR_xvf, outfile)

            # check the trajectory contents match reference universes
            for ts, ref_ts in zip(uw.trajectory, universe.trajectory):
                assert_almost_equal(
                    ts.dimensions, ref_ts.dimensions, self.prec
                )
                if pos:
                    assert_almost_equal(ts._pos, ref_ts._pos, self.prec)
                else:
                    with pytest.raises(
                        NoDataError, match="This Timestep has no"
                    ):
                        getattr(ts, "positions")
                if vel:
                    assert_almost_equal(
                        ts._velocities, ref_ts._velocities, self.prec
                    )
                else:
                    with pytest.raises(
                        NoDataError, match="This Timestep has no"
                    ):
                        getattr(ts, "velocities")
                if force:
                    assert_almost_equal(ts._forces, ref_ts._forces, self.prec)
                else:
                    with pytest.raises(
                        NoDataError, match="This Timestep has no"
                    ):
                        getattr(ts, "forces")

        # when (False, False, False)
        except ValueError:
            with pytest.raises(
                ValueError, match="At least one of positions, velocities"
            ):
                with Writer(
                    outfile,
                    universe.atoms.n_atoms,
                    positions=pos,
                    velocities=vel,
                    forces=force,
                    author="My Name",
                    author_email="my_email@asu.edu",
                ) as W:
                    for ts in universe.trajectory:
                        W.write(universe)

    def test_write_AtomGroup_with(self, universe, outfile, outtop, Writer):
        """test to write H5MD from AtomGroup"""
        ca = universe.select_atoms("protein and name CA")
        ca.write(outtop)
        with Writer(outfile, n_atoms=ca.n_atoms) as W:
            for ts in universe.trajectory:
                W.write(ca)

        uw = mda.Universe(outtop, outfile)
        caw = uw.atoms

        for orig_ts, written_ts in zip(universe.trajectory, uw.trajectory):
            assert_almost_equal(ca.positions, caw.positions, self.prec)
            assert_almost_equal(orig_ts.time, written_ts.time, self.prec)
            assert_almost_equal(
                written_ts.dimensions, orig_ts.dimensions, self.prec
            )

    @pytest.mark.parametrize("frames, n_frames", ((None, 1), ("all", 3)))
    def test_ag_write(
        self, universe, outfile, outtop, Writer, frames, n_frames
    ):
        """test to write with ag.write()"""
        ca = universe.select_atoms("protein and name CA")
        ca.write(outtop)

        ca.write(outfile, frames=frames, format="h5md")

        uw = mda.Universe(outtop, outfile)
        caw = uw.atoms

        assert_equal(n_frames, len(uw.trajectory))
        for orig_ts, written_ts in zip(universe.trajectory, uw.trajectory):
            assert_almost_equal(ca.positions, caw.positions, self.prec)
            assert_almost_equal(orig_ts.time, written_ts.time, self.prec)
            assert_almost_equal(
                written_ts.dimensions, orig_ts.dimensions, self.prec
            )

    @pytest.mark.parametrize(
        "timeunit, lengthunit, velocityunit, forceunit",
        (
            ("fs", "Angstrom", "Angstrom/ps", "kJ/(mol*Angstrom)"),
            ("s", "pm", "m/s", "Newton"),
            (
                "ps",
                "fm",
                "Angstrom/fs",
                "kcal/(mol*Angstrom)",
            ),
            ("AKMA", "nm", "Angstrom/AKMA", "kcal/(mol*Angstrom)"),
        ),
    )
    def test_write_custom_units(
        self,
        universe,
        outfile,
        Writer,
        timeunit,
        lengthunit,
        velocityunit,
        forceunit,
    ):
        with Writer(
            outfile,
            universe.atoms.n_atoms,
            lengthunit=lengthunit,
            velocityunit=velocityunit,
            forceunit=forceunit,
            timeunit=timeunit,
        ) as W:
            for ts in universe.trajectory:
                W.write(universe)

        u = mda.Universe(TPR_xvf, outfile)
        for u_unit, custom_unit in zip(
            u.trajectory.units.values(),
            (timeunit, lengthunit, velocityunit, forceunit),
        ):
            assert_equal(u_unit, custom_unit)

    @pytest.mark.parametrize(
        "timeunit, lengthunit, velocityunit, forceunit",
        (
            ("imaginary time", None, None, None),
            (None, None, "c", None),
            (
                None,
                None,
                None,
                "HUGE FORCE",
            ),
            (None, "lightyear", None, None),
        ),
    )
    def test_write_bad_units(
        self,
        universe,
        outfile,
        Writer,
        timeunit,
        lengthunit,
        velocityunit,
        forceunit,
    ):
        with pytest.raises(ValueError, match=" is not a unit recognized by"):
            with Writer(
                outfile,
                universe.atoms.n_atoms,
                lengthunit=lengthunit,
                velocityunit=velocityunit,
                forceunit=forceunit,
                timeunit=timeunit,
            ) as W:
                for ts in universe.trajectory:
                    W.write(universe)

    def test_no_units_w_convert_true(self, universe_no_units, outfile, Writer):
        # no units + convert_units = ValueError
        with pytest.raises(ValueError, match="The trajectory has no units,"):
            with Writer(outfile, universe_no_units.atoms.n_atoms) as W:
                for ts in universe_no_units.trajectory:
                    W.write(universe_no_units)

    def test_no_units_w_convert_false(
        self, universe_no_units, outfile, Writer
    ):
        with Writer(
            outfile, universe_no_units.atoms.n_atoms, convert_units=False
        ) as W:
            for ts in universe_no_units.trajectory:
                W.write(universe_no_units)

            uw = mda.Universe(TPR_xvf, outfile, convert_units=False)
            for unit in uw.trajectory.units.values():
                assert_equal(unit, None)

    @pytest.mark.parametrize("convert_units", (True, False))
    def test_convert_units(self, universe, outfile, Writer, convert_units):
        with Writer(
            outfile, universe.atoms.n_atoms, convert_units=convert_units
        ) as W:
            for ts in universe.trajectory:
                W.write(universe)

        ref_units = universe.trajectory.units.items()
        uw = mda.Universe(TPR_xvf, outfile)
        uw_units = uw.trajectory.units.items()
        for u1, u2 in zip(ref_units, uw_units):
            assert_equal(u1, u2)

    @pytest.mark.parametrize(
        "chunks", ((3, 1000, 1), (1, 1000, 3), (100, 100, 3))
    )
    def test_write_chunks(self, universe, outfile, Writer, chunks):
        with Writer(outfile, universe.atoms.n_atoms, chunks=chunks) as W:
            for ts in universe.trajectory:
                W.write(universe)

        uw = mda.Universe(TPR_xvf, outfile)
        for dset in (
            uw.trajectory._particle_group["position/value"],
            uw.trajectory._particle_group["velocity/value"],
            uw.trajectory._particle_group["force/value"],
        ):
            assert_equal(dset.chunks, chunks)

        for ts1, ts2 in zip(universe.trajectory, uw.trajectory):
            assert_equal(ts1.positions, ts2.positions)
            assert_equal(ts1.velocities, ts2.velocities)
            assert_equal(ts1.forces, ts2.forces)

    def test_write_chunks_with_nframes(self, universe, outfile, Writer):
        n_atoms = universe.atoms.n_atoms
        n_frames = universe.trajectory.n_frames
        with Writer(outfile, n_atoms=n_atoms, n_frames=n_frames) as W:
            for ts in universe.trajectory:
                W.write(universe)

        uw = mda.Universe(TPR_xvf, outfile)
        for dset in (
            uw.trajectory._particle_group["position/value"],
            uw.trajectory._particle_group["velocity/value"],
            uw.trajectory._particle_group["force/value"],
        ):
            assert_equal(dset.chunks, (1, n_atoms, 3))

        for ts1, ts2 in zip(universe.trajectory, uw.trajectory):
            assert_equal(ts1.positions, ts2.positions)
            assert_equal(ts1.velocities, ts2.velocities)
            assert_equal(ts1.forces, ts2.forces)

    def test_write_contiguous1(self, universe, Writer, outfile):
        n_atoms = universe.atoms.n_atoms
        n_frames = len(universe.trajectory)
        with Writer(
            outfile, n_atoms=n_atoms, n_frames=n_frames, chunks=False
        ) as W:
            for ts in universe.trajectory:
                W.write(universe)

        uw = mda.Universe(TPR_xvf, outfile)
        for dset in (
            uw.trajectory._particle_group["position/value"],
            uw.trajectory._particle_group["velocity/value"],
            uw.trajectory._particle_group["force/value"],
        ):
            assert_equal(dset.chunks, None)

    def test_write_contiguous2(self, universe, Writer, outfile):
        ag = universe.select_atoms("all")
        n_frames = len(ag.universe.trajectory)
        ag.write(outfile, frames="all", n_frames=n_frames, chunks=False)

        uw = mda.Universe(TPR_xvf, outfile)
        for dset in (
            uw.trajectory._particle_group["position/value"],
            uw.trajectory._particle_group["velocity/value"],
            uw.trajectory._particle_group["force/value"],
        ):
            assert_equal(dset.chunks, None)

    @pytest.mark.parametrize(
        "filter, opts", (("gzip", 1), ("gzip", 9), ("lzf", None))
    )
    def test_write_with_compression(
        self, universe, outfile, Writer, filter, opts
    ):
        with Writer(
            outfile,
            universe.atoms.n_atoms,
            compression=filter,
            compression_opts=opts,
        ) as W:
            for ts in universe.trajectory:
                W.write(universe)

        uw = mda.Universe(TPR_xvf, outfile)
        dset = uw.trajectory._particle_group["position/value"]
        assert_equal(dset.compression, filter)
        assert_equal(dset.compression_opts, opts)

    @pytest.mark.xfail(
        os.name == "nt", reason="occasional PermissionError on windows"
    )
    @pytest.mark.parametrize("driver", ("core", "stdio"))
    def test_write_with_drivers(self, universe, outfile, Writer, driver):
        with Writer(outfile, universe.atoms.n_atoms, driver=driver) as W:
            for ts in universe.trajectory:
                W.write(universe)

        uw = mda.Universe(TPR_xvf, outfile, driver=driver)
        file = uw.trajectory._file
        assert_equal(file.driver, driver)

    def test_parallel_disabled(self, universe, Writer, outfile, driver="mpio"):
        with pytest.raises(
            ValueError, match="H5MDWriter: parallel writing with MPI I/O "
        ):
            with Writer(outfile, universe.atoms.n_atoms, driver=driver) as W:
                for ts in universe.trajectory:
                    W.write(universe)

    def test_timestep_not_modified_by_writer(self, universe, Writer, outfile):
        trj = universe.trajectory
        ts = trj.ts

        trj[-1]  # last timestep (so that time != 0)
        x = ts._pos.copy()
        time = ts.time

        with Writer(outfile, trj.n_atoms) as W:
            # last timestep (so that time != 0) (say it again, just in case...)
            trj[-1]
            W.write(universe)

        assert_equal(
            ts._pos,
            x,
            err_msg="Positions in Timestep were modified by writer.",
        )
        assert_equal(
            ts.time, time, err_msg="Time in Timestep was modified by writer."
        )


class TestH5PYNotInstalled(object):
    """Tests RuntimeErrors when h5py not installed"""

    @pytest.fixture(autouse=True)
    def block_h5py(self, monkeypatch):
        monkeypatch.setattr(
            sys.modules["MDAnalysis.coordinates.H5MD"], "HAS_H5PY", False
        )

    @pytest.fixture()
    def Writer(self):
        return mda.coordinates.H5MD.H5MDWriter

    @pytest.fixture()
    def outfile(self, tmpdir):
        return str(tmpdir) + "h5md-writer-test.h5md"

    def test_reader_no_h5py(self):
        with pytest.raises(RuntimeError, match="Please install h5py"):
            u = mda.Universe(TPR_xvf, H5MD_xvf)

    def test_writer_no_h5py(self, Writer, outfile):
        u = mda.Universe(TPR_xvf, TRR_xvf)
        with pytest.raises(
            RuntimeError, match="H5MDWriter: Please install h5py"
        ):
            with Writer(outfile, u.atoms.n_atoms) as W:
                for ts in u.trajectory:
                    W.write(universe)


@pytest.mark.skipif(not HAS_H5PY, reason="h5py not installed")
class TestH5MDReaderWithObservables(object):
    """Read H5MD file with 'observables/atoms/energy'."""

    prec = 3
    ext = "h5md"

    def test_read_h5md_issue4598(self):
        """Read a H5MD file with observables.

        The reader will ignore the 'observables/atoms/energy'.
        """

        u = mda.Universe.empty(n_atoms=108, trajectory=True)
        reader = H5MDReader(H5MD_energy)
        u.trajectory = reader
        for ts in u.trajectory:
            assert "atoms/energy" in ts.data

    def test_read_h5md_malformed(self):
        """Try reading an H5MD file with malformed observables."""

        with pytest.raises(ValueError):
            H5MDReader(H5MD_malformed)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
#

"""Fast distance array computation --- :mod:`MDAnalysis.lib.distances`
===================================================================

Fast C-routines to calculate arrays of distances or angles from coordinate
arrays. Distance functions can accept a NumPy :class:`np.ndarray` or an
:class:`~MDAnalysis.core.groups.AtomGroup`. Many of the functions also exist
in parallel versions, which typically provide higher performance than the
serial code. The boolean attribute `MDAnalysis.lib.distances.USED_OPENMP` can
be checked to see if OpenMP was used in the compilation of MDAnalysis.

Selection of acceleration ("backend")
-------------------------------------

All functions take the optional keyword `backend`, which determines the type of
acceleration. Currently, the following choices are implemented (`backend` is
case-insensitive):

.. Table:: Available *backends* for accelerated distance functions.

   ========== ========================= ======================================
   *backend*  module                    description
   ========== ========================= ======================================
   "serial"   :mod:`c_distances`        serial implementation in C/Cython

   "OpenMP"   :mod:`c_distances_openmp` parallel implementation in C/Cython
                                        with OpenMP

   "distopia"  `_distopia`              SIMD-accelerated implementation
                                        with the `distopia`_ library
   ========== ========================= ======================================


Use of the distopia library
---------------------------

MDAnalysis has developed a standalone library, `distopia`_ for accelerating
the distance functions in this module using explicit SIMD vectorisation.
This can provide many-fold speedups in calculating distances. Distopia is
under active development and as such only a selection of functions in this
module are covered. Consult the following table to see if the function
you wish to use is covered by distopia. For more information see the
`distopia documentation`_.

.. table:: Functions available using the `distopia`_ backend.
   :align: center

   +-------------------------------------------------------+
   | Functions                                             |
   +=======================================================+
   | :func:`MDAnalysis.lib.distances.calc_bonds`           |
   +-------------------------------------------------------+
   | :func:`MDAnalysis.lib.distances.calc_angles`          |
   +-------------------------------------------------------+
   | :func:`MDAnalysis.lib.distances.calc_dihedrals`       |
   +-------------------------------------------------------+
   | :func:`MDAnalysis.lib.distances.distance_array`       |
   +-------------------------------------------------------+
   | :func:`MDAnalysis.lib.distances.self_distance_array`  |
   +-------------------------------------------------------+

If `distopia`_ is installed, the functions in this table will accept the key
'distopia' for the `backend` keyword argument. The variable
:data:`HAS_DISTOPIA` is set to ``True`` if distopia is available.

If the distopia backend is selected the `distopia` library will be used to
calculate the distances. Note that for functions listed in this table
**distopia is not the default backend and must be explicitly selected.**


.. Note::
    Due to the use of Instruction Set Architecture (`ISA`_) specific SIMD
    intrinsics in distopia via `HWY`_, the precision of your results may
    depend on the ISA available on your machine. However, in all tested cases
    distopia satisfied the accuracy thresholds used to the functions in this
    module. Please document any issues you encounter with distopia's accuracy
    in the `relevant distopia issue`_ on the MDAnalysis GitHub repository.

.. _distopia: https://github.com/MDAnalysis/distopia
.. _distopia documentation: https://www.mdanalysis.org/distopia
.. _ISA: https://en.wikipedia.org/wiki/Instruction_set_architecture
.. _HWY: https://github.com/google/highway
.. _relevant distopia issue: https://github.com/MDAnalysis/mdanalysis/issues/3915

.. versionadded:: 0.13.0
.. versionchanged:: 2.3.0
   Distance functions can now accept an
   :class:`~MDAnalysis.core.groups.AtomGroup` or an :class:`np.ndarray`
.. versionchanged:: 2.5.0
   Interface to the `distopia`_ package added.
.. versionchanged:: 2.9.0
   Distopia support greatly expanded (with distopia ≥ 0.4.0).

Constants
---------
.. data:: HAS_DISTOPIA

   This variable is ``True`` if the :mod:`distopia` package has been
   installed and is available as a `backend`. Otherwise it is
   ``False``.

Functions
---------
.. autofunction:: distance_array
.. autofunction:: self_distance_array
.. autofunction:: capped_distance
.. autofunction:: self_capped_distance
.. autofunction:: calc_bonds
.. autofunction:: calc_angles
.. autofunction:: calc_dihedrals
.. autofunction:: apply_PBC
.. autofunction:: transform_RtoS
.. autofunction:: transform_StoR
.. autofunction:: augment_coordinates(coordinates, box, r)
.. autofunction:: undo_augment(results, translation, nreal)
.. autofunction:: minimize_vectors(vectors, box)

"""
import numpy as np
import numpy.typing as npt

from typing import Union, Optional, Callable
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from ..core.groups import AtomGroup
from .util import check_coords, check_box
from .mdamath import triclinic_vectors
from ._augment import augment_coordinates, undo_augment
from .nsgrid import FastNS
from .c_distances import _minimize_vectors_ortho, _minimize_vectors_triclinic
from ._distopia import HAS_DISTOPIA


# hack to select backend with backend=<backend> kwarg. Note that
# the cython parallel code (prange) in parallel.distances is
# independent from the OpenMP code
import importlib

_distances = {}
_distances["serial"] = importlib.import_module(
    ".c_distances", package="MDAnalysis.lib"
)
try:
    _distances["openmp"] = importlib.import_module(
        ".c_distances_openmp", package="MDAnalysis.lib"
    )
except ImportError:
    pass

if HAS_DISTOPIA:
    _distances["distopia"] = importlib.import_module(
        "._distopia", package="MDAnalysis.lib"
    )
del importlib


def _run(
    funcname: str,
    args: Optional[tuple] = None,
    kwargs: Optional[dict] = None,
    backend: str = "serial",
) -> Callable:
    """Helper function to select a backend function `funcname`."""
    args = args if args is not None else tuple()
    kwargs = kwargs if kwargs is not None else dict()
    backend = backend.lower()
    try:
        func = getattr(_distances[backend], funcname)
    except KeyError:
        errmsg = (
            f"Function {funcname} not available with backend {backend} "
            f"try one of: {_distances.keys()}"
        )
        raise ValueError(errmsg) from None
    return func(*args, **kwargs)


# serial versions are always available (and are typically used within
# the core and topology modules)
from .c_distances import (
    _UINT64_MAX,
    calc_distance_array,
    calc_distance_array_ortho,
    calc_distance_array_triclinic,
    calc_self_distance_array,
    calc_self_distance_array_ortho,
    calc_self_distance_array_triclinic,
    coord_transform,
    calc_bond_distance,
    calc_bond_distance_ortho,
    calc_bond_distance_triclinic,
    calc_angle,
    calc_angle_ortho,
    calc_angle_triclinic,
    calc_dihedral,
    calc_dihedral_ortho,
    calc_dihedral_triclinic,
    ortho_pbc,
    triclinic_pbc,
)

from .c_distances_openmp import OPENMP_ENABLED as USED_OPENMP


def _check_result_array(
    result: Optional[npt.NDArray], shape: tuple
) -> npt.NDArray:
    """Check if the result array is ok to use.

    The `result` array must meet the following requirements:
      * Must have a shape equal to `shape`.
      * Its dtype must be ``numpy.float64``.

    Paramaters
    ----------
    result : numpy.ndarray or None
        The result array to check. If `result` is `None``, a newly created
        array of correct shape and dtype ``numpy.float64`` will be returned.
    shape : tuple
        The shape expected for the `result` array.

    Returns
    -------
    result : numpy.ndarray (``dtype=numpy.float64``, ``shape=shape``)
        The input array or a newly created array if the input was ``None``.

    Raises
    ------
    ValueError
        If `result` is of incorrect shape.
    TypeError
        If the dtype of `result` is not ``numpy.float64``.
    """
    if result is None:
        return np.zeros(shape, dtype=np.float64)
    if result.shape != shape:
        raise ValueError(
            "Result array has incorrect shape, should be {0}, got "
            "{1}.".format(shape, result.shape)
        )
    if result.dtype != np.float64:
        raise TypeError(
            "Result array must be of type numpy.float64, got {}."
            "".format(result.dtype)
        )
    # The following two lines would break a lot of tests. WHY?!
    #    if not coords.flags['C_CONTIGUOUS']:
    #        raise ValueError("{0} is not C-contiguous.".format(desc))
    return result


@check_coords(
    "reference",
    "configuration",
    reduce_result_if_single=False,
    check_lengths_match=False,
    allow_atomgroup=True,
)
def distance_array(
    reference: Union[npt.NDArray, "AtomGroup"],
    configuration: Union[npt.NDArray, "AtomGroup"],
    box: Optional[npt.NDArray] = None,
    result: Optional[npt.NDArray] = None,
    backend: str = "serial",
) -> npt.NDArray:
    """Calculate all possible distances between a reference set and another
    configuration.

    If there are ``n`` positions in `reference` and ``m`` positions in
    `configuration`, a distance array of shape ``(n, m)`` will be computed.

    If the optional argument `box` is supplied, the minimum image convention is
    applied when calculating distances. Either orthogonal or triclinic boxes are
    supported.

    If a 2D numpy array of dtype ``numpy.float64`` with the shape ``(n, m)``
    is provided in `result`, then this preallocated array is filled. This can
    speed up calculations.

    Parameters
    ----------
    reference :numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Reference coordinate array of shape ``(3,)`` or ``(n, 3)`` (dtype is
        arbitrary, will be converted to ``numpy.float32`` internally). Also
        accepts an :class:`~MDAnalysis.core.groups.AtomGroup`.
    configuration : numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Configuration coordinate array of shape ``(3,)`` or ``(m, 3)`` (dtype is
        arbitrary, will be converted to ``numpy.float32`` internally). Also
        accepts an :class:`~MDAnalysis.core.groups.AtomGroup`.
    box : array_like, optional
        The unitcell dimensions of the system, which can be orthogonal or
        triclinic and must be provided in the same format as returned by
        :attr:`MDAnalysis.coordinates.timestep.Timestep.dimensions`:
        ``[lx, ly, lz, alpha, beta, gamma]``.
    result : numpy.ndarray, optional
        Preallocated result array which must have the shape ``(n, m)`` and dtype
        ``numpy.float64``.
        Avoids creating the array which saves time when the function
        is called repeatedly.
    backend : {'serial', 'OpenMP', 'distopia'}, optional
        Keyword selecting the type of acceleration.

    Returns
    -------
    d : numpy.ndarray (``dtype=numpy.float64``, ``shape=(n, m)``)
        Array containing the distances ``d[i,j]`` between reference coordinates
        ``i`` and configuration coordinates ``j``.


    .. versionchanged:: 0.13.0
       Added *backend* keyword.
    .. versionchanged:: 0.19.0
       Internal dtype conversion of input coordinates to ``numpy.float32``.
       Now also accepts single coordinates as input.
    .. versionchanged:: 2.3.0
       Can now accept an :class:`~MDAnalysis.core.groups.AtomGroup` as an
       argument in any position and checks inputs using type hinting.
    .. versionchanged:: 2.9.0
       Added support for the `distopia` backend.
    """
    confnum = configuration.shape[0]
    refnum = reference.shape[0]

    # check resulting array will not overflow UINT64_MAX
    if refnum * confnum > _UINT64_MAX:
        raise ValueError(
            f"Size of resulting array {refnum * confnum} elements"
            " larger than size of maximum integer"
        )

    distances = _check_result_array(result, (refnum, confnum))
    if len(distances) == 0:
        return distances

    if backend == "distopia":
        # distopia requires that all the input arrays are the same type,
        # while MDAnalysis allows for mixed types, this should be changed
        # pre 3.0.0 release see issue #3707
        distances = distances.astype(np.float32)
        box = np.asarray(box).astype(np.float32) if box is not None else None

    if box is not None:
        boxtype, box = check_box(box)
        if boxtype == "ortho":
            _run(
                "calc_distance_array_ortho",
                args=(reference, configuration, box, distances),
                backend=backend,
            )
        else:
            _run(
                "calc_distance_array_triclinic",
                args=(reference, configuration, box, distances),
                backend=backend,
            )
    else:
        _run(
            "calc_distance_array",
            args=(reference, configuration, distances),
            backend=backend,
        )

    if backend == "distopia":
        # mda expects the result to be in float64, so we need to convert it back
        # to float64, change for 3.0, see #3707
        distances = distances.astype(np.float64)
        if result is not None:
            result[:] = distances

    return distances


@check_coords("reference", reduce_result_if_single=False, allow_atomgroup=True)
def self_distance_array(
    reference: Union[npt.NDArray, "AtomGroup"],
    box: Optional[npt.NDArray] = None,
    result: Optional[npt.NDArray] = None,
    backend: str = "serial",
) -> npt.NDArray:
    """Calculate all possible distances within a configuration `reference`.

    If the optional argument `box` is supplied, the minimum image convention is
    applied when calculating distances. Either orthogonal or triclinic boxes are
    supported.

    If a 1D numpy array of dtype ``numpy.float64`` with the shape
    ``(n*(n-1)/2,)`` is provided in `result`, then this preallocated array is
    filled. This can speed up calculations.

    Parameters
    ----------
    reference : numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Reference coordinate array of shape ``(3,)`` or ``(n, 3)`` (dtype is
        arbitrary, will be converted to ``numpy.float32`` internally). Also
        accepts an :class:`~MDAnalysis.core.groups.AtomGroup`.
    box : array_like, optional
        The unitcell dimensions of the system, which can be orthogonal or
        triclinic and must be provided in the same format as returned by
        :attr:`MDAnalysis.coordinates.timestep.Timestep.dimensions`:
        ``[lx, ly, lz, alpha, beta, gamma]``.
    result : numpy.ndarray, optional
        Preallocated result array which must have the shape ``(n*(n-1)/2,)`` and
        dtype ``numpy.float64``. Avoids creating the array which saves time when
        the function is called repeatedly.
    backend : {'serial', 'OpenMP', 'distopia'}, optional
        Keyword selecting the type of acceleration.

    Returns
    -------
    d : numpy.ndarray (``dtype=numpy.float64``, ``shape=(n*(n-1)/2,)``)
        Array containing the distances ``dist[i,j]`` between reference
        coordinates ``i`` and ``j`` at position ``d[k]``. Loop through ``d``:

        .. code-block:: python

            for i in range(n):
                for j in range(i + 1, n):
                    dist[i, j] = dist[j, i] = d[k]
                    k += 1

    .. versionchanged:: 0.13.0
       Added *backend* keyword.
    .. versionchanged:: 0.19.0
       Internal dtype conversion of input coordinates to ``numpy.float32``.
    .. versionchanged:: 2.3.0
       Can now accept an :class:`~MDAnalysis.core.groups.AtomGroup` as an
       argument in any position and checks inputs using type hinting.
    .. versionchanged:: 2.9.0
       Added support for the `distopia` backend.
    """
    refnum = reference.shape[0]
    distnum = refnum * (refnum - 1) // 2
    # check resulting array will not overflow UINT64_MAX
    if distnum > _UINT64_MAX:
        raise ValueError(
            f"Size of resulting array {distnum} elements larger"
            " than size of maximum integer"
        )

    distances = _check_result_array(result, (distnum,))
    if len(distances) == 0:
        return distances

    if backend == "distopia":
        # distopia requires that all the input arrays are the same type,
        # while MDAnalysis allows for mixed types, this should be changed
        # pre 3.0.0 release see issue #3707
        distances = distances.astype(np.float32)
        box = np.asarray(box).astype(np.float32) if box is not None else None

    if box is not None:
        boxtype, box = check_box(box)
        if boxtype == "ortho":
            _run(
                "calc_self_distance_array_ortho",
                args=(reference, box, distances),
                backend=backend,
            )
        else:
            _run(
                "calc_self_distance_array_triclinic",
                args=(reference, box, distances),
                backend=backend,
            )
    else:
        _run(
            "calc_self_distance_array",
            args=(reference, distances),
            backend=backend,
        )

    if backend == "distopia":
        # mda expects the result to be in float64, so we need to convert it back
        # to float64, change for 3.0, see #3707
        distances = distances.astype(np.float64)
        if result is not None:
            result[:] = distances

    return distances


@check_coords(
    "reference",
    "configuration",
    enforce_copy=False,
    reduce_result_if_single=False,
    check_lengths_match=False,
    allow_atomgroup=True,
)
def capped_distance(
    reference: Union[npt.NDArray, "AtomGroup"],
    configuration: Union[npt.NDArray, "AtomGroup"],
    max_cutoff: float,
    min_cutoff: Optional[float] = None,
    box: Optional[npt.NDArray] = None,
    method: Optional[str] = None,
    return_distances: Optional[bool] = True,
    backend: Optional[str] = "serial",
):
    """Calculates pairs of indices corresponding to entries in the `reference`
    and `configuration` arrays which are separated by a distance lying within
    the specified cutoff(s). Optionally, these distances can be returned as
    well.

    If the optional argument `box` is supplied, the minimum image convention is
    applied when calculating distances. Either orthogonal or triclinic boxes are
    supported.

    An automatic guessing of the optimal method to calculate the distances is
    included in the function. An optional keyword for the method is also
    provided. Users can enforce a particular method with this functionality.
    Currently brute force, grid search, and periodic KDtree methods are
    implemented.

    Parameters
    -----------
    reference : numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Reference coordinate array with shape ``(3,)`` or ``(n, 3)``. Also
        accepts an :class:`~MDAnalysis.core.groups.AtomGroup`.
    configuration : numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Configuration coordinate array with shape ``(3,)`` or ``(m, 3)``. Also
        accepts an :class:`~MDAnalysis.core.groups.AtomGroup`.
    max_cutoff : float
        Maximum cutoff distance between the reference and configuration.
    min_cutoff : float, optional
        Minimum cutoff distance between reference and configuration.
    box : array_like, optional
        The unitcell dimensions of the system, which can be orthogonal or
        triclinic and must be provided in the same format as returned by
        :attr:`MDAnalysis.coordinates.timestep.Timestep.dimensions`:
        ``[lx, ly, lz, alpha, beta, gamma]``.
    method : {'bruteforce', 'nsgrid', 'pkdtree'}, optional
        Keyword to override the automatic guessing of the employed search
        method.
    return_distances : bool, optional
        If set to ``True``, distances will also be returned.
    backend : {'serial', 'OpenMP', 'distopia'}, optional
        Keyword selecting the type of acceleration.

    Returns
    -------
    pairs : numpy.ndarray (``dtype=numpy.int64``, ``shape=(n_pairs, 2)``)
        Pairs of indices, corresponding to coordinates in the `reference` and
        `configuration` arrays such that the distance between them lies within
        the interval (`min_cutoff`, `max_cutoff`].
        Each row in `pairs` is an index pair ``[i, j]`` corresponding to the
        ``i``-th coordinate in `reference` and the ``j``-th coordinate in
        `configuration`.
    distances : numpy.ndarray (``dtype=numpy.float64``, ``shape=(n_pairs,)``), optional
        Distances corresponding to each pair of indices. Only returned if
        `return_distances` is ``True``. ``distances[k]`` corresponds to the
        ``k``-th pair returned in `pairs` and gives the distance between the
        coordinates ``reference[pairs[k, 0]]`` and
        ``configuration[pairs[k, 1]]``.

        .. code-block:: python

            pairs, distances = capped_distances(reference, configuration,
                                                max_cutoff, return_distances=True)
            for k, [i, j] in enumerate(pairs):
                coord1 = reference[i]
                coord2 = configuration[j]
                distance = distances[k]

    See Also
    --------
    distance_array
    MDAnalysis.lib.pkdtree.PeriodicKDTree.search
    MDAnalysis.lib.nsgrid.FastNS.search


    .. versionchanged:: 1.0.1
       nsgrid was temporarily removed and replaced with pkdtree due to issues
       relating to its reliability and accuracy (Issues #2919, #2229, #2345,
       #2670, #2930)
    .. versionchanged:: 1.0.2
       nsgrid enabled again
    .. versionchanged:: 2.3.0
       Can now accept an :class:`~MDAnalysis.core.groups.AtomGroup` as an
       argument in any position and checks inputs using type hinting.
    .. versionchanged:: 2.10.0
       Added the "backend" argument to select the type of acceleration of
       the distance calculations.
    """
    if box is not None:
        box = np.asarray(box, dtype=np.float32)
        if box.shape[0] != 6:
            raise ValueError(
                "Box Argument is of incompatible type. The "
                "dimension should be either None or of the form "
                "[lx, ly, lz, alpha, beta, gamma]"
            )

    # The check_coords decorator made sure that reference and configuration
    # are arrays of positions. Mypy does not know about that so we have to
    # tell it.
    reference_positions: npt.NDArray = reference  # type: ignore
    configuration_positions: npt.NDArray = configuration  # type: ignore
    function = _determine_method(
        reference_positions,
        configuration_positions,
        max_cutoff,
        min_cutoff=min_cutoff,
        box=box,
        method=method,
    )

    if function.__name__ == "_nsgrid_capped":
        return function(
            reference,
            configuration,
            max_cutoff,
            min_cutoff=min_cutoff,
            box=box,
            return_distances=return_distances,
        )
    else:
        return function(
            reference,
            configuration,
            max_cutoff,
            min_cutoff=min_cutoff,
            box=box,
            return_distances=return_distances,
            backend=backend,
        )


def _determine_method(
    reference: npt.NDArray,
    configuration: npt.NDArray,
    max_cutoff: float,
    min_cutoff: Optional[float] = None,
    box: Optional[npt.NDArray] = None,
    method: Optional[str] = None,
) -> Callable:
    """Guesses the fastest method for capped distance calculations based on the
    size of the coordinate sets and the relative size of the target volume.

    Parameters
    ----------
    reference : numpy.ndarray
        Reference coordinate array with shape ``(3,)`` or ``(n, 3)``.
    configuration : numpy.ndarray
        Configuration coordinate array with shape ``(3,)`` or ``(m, 3)``.
    max_cutoff : float
        Maximum cutoff distance between `reference` and `configuration`
        coordinates.
    min_cutoff : float, optional
        Minimum cutoff distance between `reference` and `configuration`
        coordinates.
    box : numpy.ndarray
        The unitcell dimensions of the system, which can be orthogonal or
        triclinic and must be provided in the same format as returned by
        :attr:`MDAnalysis.coordinates.timestep.Timestep.dimensions`:
        ``[lx, ly, lz, alpha, beta, gamma]``.
    method : {'bruteforce', 'nsgrid', 'pkdtree'}, optional
        Keyword to override the automatic guessing of the employed search
        method.

    Returns
    -------
    function : callable
        The function implementing the guessed (or deliberatly chosen) method.


    .. versionchanged:: 1.0.1
       nsgrid was temporarily removed and replaced with pkdtree due to issues
       relating to its reliability and accuracy (Issues #2919, #2229, #2345,
       #2670, #2930)
    .. versionchanged:: 1.1.0
       enabled nsgrid again
    """
    methods = {
        "bruteforce": _bruteforce_capped,
        "pkdtree": _pkdtree_capped,
        "nsgrid": _nsgrid_capped,
    }

    if method is not None:
        return methods[method.lower()]

    if len(reference) < 10 or len(configuration) < 10:
        return methods["bruteforce"]
    elif len(reference) * len(configuration) >= 1e8:
        # CAUTION : for large datasets, shouldnt go into 'bruteforce'
        # in any case. Arbitrary number, but can be characterized
        return methods["nsgrid"]
    else:
        if box is None:
            min_dim = np.array(
                [reference.min(axis=0), configuration.min(axis=0)]
            )
            max_dim = np.array(
                [reference.max(axis=0), configuration.max(axis=0)]
            )
            size = max_dim.max(axis=0) - min_dim.min(axis=0)
        elif np.all(box[3:] == 90.0):
            size = box[:3]
        else:
            tribox = triclinic_vectors(box)
            size = tribox.max(axis=0) - tribox.min(axis=0)
        if np.any(max_cutoff > 0.3 * size):
            return methods["bruteforce"]
        else:
            return methods["nsgrid"]


@check_coords(
    "reference",
    "configuration",
    enforce_copy=False,
    reduce_result_if_single=False,
    check_lengths_match=False,
    allow_atomgroup=True,
)
def _bruteforce_capped(
    reference: Union[npt.NDArray, "AtomGroup"],
    configuration: Union[npt.NDArray, "AtomGroup"],
    max_cutoff: float,
    min_cutoff: Optional[float] = None,
    box: Optional[npt.NDArray] = None,
    return_distances: Optional[bool] = True,
    backend: Optional[str] = "serial",
):
    """Capped distance evaluations using a brute force method.

    Computes and returns an array containing pairs of indices corresponding to
    entries in the `reference` and `configuration` arrays which are separated by
    a distance lying within the specified cutoff(s). Employs naive distance
    computations (brute force) to find relevant distances.

    Optionally, these distances can be returned as well.

    If the optional argument `box` is supplied, the minimum image convention is
    applied when calculating distances. Either orthogonal or triclinic boxes are
    supported.

    Parameters
    ----------
    reference : numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Reference coordinate array with shape ``(3,)`` or ``(n, 3)`` (dtype will
        be converted to ``numpy.float32`` internally). Also
        accepts an :class:`~MDAnalysis.core.groups.AtomGroup`.
    configuration : array or :class:`~MDAnalysis.core.groups.AtomGroup`
        Configuration coordinate array with shape ``(3,)`` or ``(m, 3)`` (dtype
        will be converted to ``numpy.float32`` internally). Also
        accepts an :class:`~MDAnalysis.core.groups.AtomGroup`.
    max_cutoff : float
        Maximum cutoff distance between `reference` and `configuration`
        coordinates.
    min_cutoff : float, optional
        Minimum cutoff distance between `reference` and `configuration`
        coordinates.
    box : numpy.ndarray, optional
        The unitcell dimensions of the system, which can be orthogonal or
        triclinic and must be provided in the same format as returned by
        :attr:`MDAnalysis.coordinates.timestep.Timestep.dimensions`:
        ``[lx, ly, lz, alpha, beta, gamma]``.
    return_distances : bool, optional
        If set to ``True``, distances will also be returned.
    backend : {'serial', 'OpenMP', 'distopia'}, optional
        Keyword selecting the type of acceleration.

    Returns
    -------
    pairs : numpy.ndarray (``dtype=numpy.int64``, ``shape=(n_pairs, 2)``)
        Pairs of indices, corresponding to coordinates in the `reference` and
        `configuration` arrays such that the distance between them lies within
        the interval (`min_cutoff`, `max_cutoff`].
        Each row in `pairs` is an index pair ``[i, j]`` corresponding to the
        ``i``-th coordinate in `reference` and the ``j``-th coordinate in
        `configuration`.
    distances : numpy.ndarray (``dtype=numpy.float64``, ``shape=(n_pairs,)``), optional
        Distances corresponding to each pair of indices. Only returned if
        `return_distances` is ``True``. ``distances[k]`` corresponds to the
        ``k``-th pair returned in `pairs` and gives the distance between the
        coordinates ``reference[pairs[k, 0]]`` and
        ``configuration[pairs[k, 1]]``.

    .. versionchanged:: 2.3.0
       Can now accept an :class:`~MDAnalysis.core.groups.AtomGroup` as an
       argument in any position and checks inputs using type hinting.
    .. versionchanged:: 2.10.0
       Added the "backend" argument to select the type of acceleration of
       the distance calculations.
    """
    # Default return values (will be overwritten only if pairs are found):
    pairs = np.empty((0, 2), dtype=np.intp)
    distances = np.empty((0,), dtype=np.float64)

    if len(reference) > 0 and len(configuration) > 0:
        _distances = distance_array(
            reference, configuration, box=box, backend=backend
        )
        if min_cutoff is not None:
            mask = np.where(
                (_distances <= max_cutoff) & (_distances > min_cutoff)
            )
        else:
            mask = np.where((_distances <= max_cutoff))
        if mask[0].size > 0:
            pairs = np.c_[mask[0], mask[1]]
            if return_distances:
                distances = _distances[mask]

    if return_distances:
        return pairs, distances
    else:
        return pairs


@check_coords(
    "reference",
    "configuration",
    enforce_copy=False,
    reduce_result_if_single=False,
    check_lengths_match=False,
    allow_atomgroup=True,
)
def _pkdtree_capped(
    reference: Union[npt.NDArray, "AtomGroup"],
    configuration: Union[npt.NDArray, "AtomGroup"],
    max_cutoff: float,
    min_cutoff: Optional[float] = None,
    box: Optional[npt.NDArray] = None,
    return_distances: Optional[bool] = True,
    backend: Optional[str] = "serial",
):
    """Capped distance evaluations using a KDtree method.

    Computes and returns an array containing pairs of indices corresponding to
    entries in the `reference` and `configuration` arrays which are separated by
    a distance lying within the specified cutoff(s). Employs a (periodic) KDtree
    algorithm to find relevant distances.

    Optionally, these distances can be returned as well.

    If the optional argument `box` is supplied, the minimum image convention is
    applied when calculating distances. Either orthogonal or triclinic boxes are
    supported.

    Parameters
    ----------
    reference : numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Reference coordinate array with shape ``(3,)`` or ``(n, 3)`` (dtype will
        be converted to ``numpy.float32`` internally). Also
        accepts an :class:`~MDAnalysis.core.groups.AtomGroup`.
    configuration : numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Configuration coordinate array with shape ``(3,)`` or ``(m, 3)`` (dtype
        will be converted to ``numpy.float32`` internally). Also
        accepts an :class:`~MDAnalysis.core.groups.AtomGroup`.
    max_cutoff : float
        Maximum cutoff distance between `reference` and `configuration`
        coordinates.
    min_cutoff : float, optional
        Minimum cutoff distance between `reference` and `configuration`
        coordinates.
    box : numpy.ndarray, optional
        The unitcell dimensions of the system, which can be orthogonal or
        triclinic and must be provided in the same format as returned by
        :attr:`MDAnalysis.coordinates.timestep.Timestep.dimensions`:
        ``[lx, ly, lz, alpha, beta, gamma]``.
    return_distances : bool, optional
        If set to ``True``, distances will also be returned.
    backend : {'serial', 'OpenMP', 'distopia'}, optional
        Keyword selecting the type of acceleration.

    Returns
    -------
    pairs : numpy.ndarray (``dtype=numpy.int64``, ``shape=(n_pairs, 2)``)
        Pairs of indices, corresponding to coordinates in the `reference` and
        `configuration` arrays such that the distance between them lies within
        the interval (`min_cutoff`, `max_cutoff`].
        Each row in `pairs` is an index pair ``[i, j]`` corresponding to the
        ``i``-th coordinate in `reference` and the ``j``-th coordinate in
        `configuration`.
    distances : numpy.ndarray (``dtype=numpy.float64``, ``shape=(n_pairs,)``), optional
        Distances corresponding to each pair of indices. Only returned if
        `return_distances` is ``True``. ``distances[k]`` corresponds to the
        ``k``-th pair returned in `pairs` and gives the distance between the
        coordinates ``reference[pairs[k, 0]]`` and
        ``configuration[pairs[k, 1]]``.

    .. versionchanged:: 2.3.0
       Can now accept an :class:`~MDAnalysis.core.groups.AtomGroup` as an
       argument in any position and checks inputs using type hinting.
    .. versionchanged:: 2.10.0
       Added the "backend" argument to select the type of acceleration of
       the distance calculations.
    """
    from .pkdtree import (
        PeriodicKDTree,
    )  # must be here to avoid circular import

    # Default return values (will be overwritten only if pairs are found):
    pairs = np.empty((0, 2), dtype=np.intp)
    distances = np.empty((0,), dtype=np.float64)

    if len(reference) > 0 and len(configuration) > 0:
        kdtree = PeriodicKDTree(box=box)
        cut = max_cutoff if box is not None else None
        kdtree.set_coords(configuration, cutoff=cut)
        _pairs = kdtree.search_tree(reference, max_cutoff)
        if _pairs.size > 0:
            pairs = _pairs
            if return_distances or (min_cutoff is not None):
                refA, refB = pairs[:, 0], pairs[:, 1]
                distances = calc_bonds(
                    reference[refA],
                    configuration[refB],
                    box=box,
                    backend=backend,
                )
                if min_cutoff is not None:
                    mask = np.where(distances > min_cutoff)
                    pairs, distances = pairs[mask], distances[mask]

    if return_distances:
        return pairs, distances
    else:
        return pairs


@check_coords(
    "reference",
    "configuration",
    enforce_copy=False,
    reduce_result_if_single=False,
    check_lengths_match=False,
    allow_atomgroup=True,
)
def _nsgrid_capped(
    reference: Union[npt.NDArray, "AtomGroup"],
    configuration: Union[npt.NDArray, "AtomGroup"],
    max_cutoff: float,
    min_cutoff: Optional[float] = None,
    box: Optional[npt.NDArray] = None,
    return_distances: Optional[bool] = True,
):
    """Capped distance evaluations using a grid-based search method.

    Computes and returns an array containing pairs of indices corresponding to
    entries in the `reference` and `configuration` arrays which are separated by
    a distance lying within the specified cutoff(s). Employs a grid-based search
    algorithm to find relevant distances.

    Optionally, these distances can be returned as well.

    If the optional argument `box` is supplied, the minimum image convention is
    applied when calculating distances. Either orthogonal or triclinic boxes are
    supported.

    Parameters
    ----------
    reference : numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Reference coordinate array with shape ``(3,)`` or ``(n, 3)`` (dtype will
        be converted to ``numpy.float32`` internally). Also
        accepts an :class:`~MDAnalysis.core.groups.AtomGroup`.
    configuration : numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Configuration coordinate array with shape ``(3,)`` or ``(m, 3)`` (dtype
        will be converted to ``numpy.float32`` internally). Also
        accepts an :class:`~MDAnalysis.core.groups.AtomGroup`.
    max_cutoff : float
        Maximum cutoff distance between `reference` and `configuration`
        coordinates.
    min_cutoff : float, optional
        Minimum cutoff distance between `reference` and `configuration`
        coordinates.
    box : numpy.ndarray (``dtype=numpy.float64``, ``shape=(n_pairs,)``), optional
        The unitcell dimensions of the system, which can be orthogonal or
        triclinic and must be provided in the same format as returned by
        :attr:`MDAnalysis.coordinates.timestep.Timestep.dimensions`:
        ``[lx, ly, lz, alpha, beta, gamma]``.
    return_distances : bool, optional
        If set to ``True``, distances will also be returned.

    Returns
    -------
    pairs : numpy.ndarray (``dtype=numpy.int64``, ``shape=(n_pairs, 2)``)
        Pairs of indices, corresponding to coordinates in the `reference` and
        `configuration` arrays such that the distance between them lies within
        the interval (`min_cutoff`, `max_cutoff`].
        Each row in `pairs` is an index pair ``[i, j]`` corresponding to the
        ``i``-th coordinate in `reference` and the ``j``-th coordinate in
        `configuration`.
    distances : numpy.ndarray, optional
        Distances corresponding to each pair of indices. Only returned if
        `return_distances` is ``True``. ``distances[k]`` corresponds to the
        ``k``-th pair returned in `pairs` and gives the distance between the
        coordinates ``reference[pairs[k, 0]]`` and
        ``configuration[pairs[k, 1]]``.

    .. versionchanged:: 2.3.0
       Can now accept an :class:`~MDAnalysis.core.groups.AtomGroup` as an
       argument in any position and checks inputs using type hinting.
    """
    # Default return values (will be overwritten only if pairs are found):
    pairs = np.empty((0, 2), dtype=np.intp)
    distances = np.empty((0,), dtype=np.float64)

    if len(reference) > 0 and len(configuration) > 0:
        if box is None:
            # create a pseudobox
            # define the max range
            # and supply the pseudobox
            # along with only one set of coordinates
            pseudobox = np.zeros(6, dtype=np.float32)
            all_coords = np.concatenate([reference, configuration])
            lmax = all_coords.max(axis=0)
            lmin = all_coords.min(axis=0)
            # Using maximum dimension as the box size
            boxsize = (lmax - lmin).max()
            # to avoid failures for very close particles but with
            # larger cutoff
            boxsize = np.maximum(boxsize, 2 * max_cutoff)
            pseudobox[:3] = boxsize + 2.2 * max_cutoff
            pseudobox[3:] = 90.0
            shiftref, shiftconf = reference.copy(), configuration.copy()
            # Extra padding near the origin
            shiftref -= lmin - 0.1 * max_cutoff
            shiftconf -= lmin - 0.1 * max_cutoff
            gridsearch = FastNS(
                max_cutoff, shiftconf, box=pseudobox, pbc=False
            )
            results = gridsearch.search(shiftref)
        else:
            gridsearch = FastNS(max_cutoff, configuration, box=box)
            results = gridsearch.search(reference)

        pairs = results.get_pairs()
        if return_distances or (min_cutoff is not None):
            distances = results.get_pair_distances()
            if min_cutoff is not None:
                idx = distances > min_cutoff
                pairs, distances = pairs[idx], distances[idx]

    if return_distances:
        return pairs, distances
    else:
        return pairs


@check_coords(
    "reference",
    enforce_copy=False,
    reduce_result_if_single=False,
    check_lengths_match=False,
    allow_atomgroup=True,
)
def self_capped_distance(
    reference: Union[npt.NDArray, "AtomGroup"],
    max_cutoff: float,
    min_cutoff: Optional[float] = None,
    box: Optional[npt.NDArray] = None,
    method: Optional[str] = None,
    return_distances: Optional[bool] = True,
    backend: Optional[str] = "serial",
):
    """Calculates pairs of indices corresponding to entries in the `reference`
    array which are separated by a distance lying within the specified
    cutoff(s). Optionally, these distances can be returned as well.

    If the optional argument `box` is supplied, the minimum image convention is
    applied when calculating distances. Either orthogonal or triclinic boxes are
    supported.

    An automatic guessing of the optimal method to calculate the distances is
    included in the function. An optional keyword for the method is also
    provided. Users can enforce a particular method with this functionality.
    Currently brute force, grid search, and periodic KDtree methods are
    implemented.

    Parameters
    -----------
    reference : numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Reference coordinate array with shape ``(3,)`` or ``(n, 3)``. Also
        accepts an :class:`~MDAnalysis.core.groups.AtomGroup`.
    max_cutoff : float
        Maximum cutoff distance between `reference` coordinates.
    min_cutoff : float, optional
        Minimum cutoff distance between `reference` coordinates.
    box : array_like, optional
        The unitcell dimensions of the system, which can be orthogonal or
        triclinic and must be provided in the same format as returned by
        :attr:`MDAnalysis.coordinates.timestep.Timestep.dimensions`:
        ``[lx, ly, lz, alpha, beta, gamma]``.
    method : {'bruteforce', 'nsgrid', 'pkdtree'}, optional
        Keyword to override the automatic guessing of the employed search
        method.
    return_distances : bool, optional
        If set to ``True``, distances will also be returned.
    backend : {'serial', 'OpenMP', 'distopia'}, optional
        Keyword selecting the type of acceleration.

    Returns
    -------
    pairs : numpy.ndarray (``dtype=numpy.int64``, ``shape=(n_pairs, 2)``)
        Pairs of indices, corresponding to coordinates in the `reference` array
        such that the distance between them lies within the interval
        (`min_cutoff`, `max_cutoff`].
        Each row in `pairs` is an index pair ``[i, j]`` corresponding to the
        ``i``-th and the ``j``-th coordinate in `reference`.
    distances : numpy.ndarray (``dtype=numpy.float64``, ``shape=(n_pairs,)``)
        Distances corresponding to each pair of indices. Only returned if
        `return_distances` is ``True``. ``distances[k]`` corresponds to the
        ``k``-th pair returned in `pairs` and gives the distance between the
        coordinates ``reference[pairs[k, 0]]`` and ``reference[pairs[k, 1]]``.

        .. code-block:: python

            pairs, distances = self_capped_distances(reference, max_cutoff,
                                                     return_distances=True)
            for k, [i, j] in enumerate(pairs):
                coord1 = reference[i]
                coord2 = reference[j]
                distance = distances[k]


    Note
    -----
    Currently supports brute force, grid-based, and periodic KDtree search
    methods.

    See Also
    --------
    self_distance_array
    MDAnalysis.lib.pkdtree.PeriodicKDTree.search
    MDAnalysis.lib.nsgrid.FastNS.self_search


    .. versionchanged:: 0.20.0
       Added `return_distances` keyword.
    .. versionchanged:: 1.0.1
       nsgrid was temporarily removed and replaced with pkdtree due to issues
       relating to its reliability and accuracy (Issues #2919, #2229, #2345,
       #2670, #2930)
    .. versionchanged:: 1.0.2
       enabled nsgrid again
    .. versionchanged:: 2.3.0
       Can now accept an :class:`~MDAnalysis.core.groups.AtomGroup` as an
       argument in any position and checks inputs using type hinting.
    .. versionchanged:: 2.10.0
       Added the "backend" argument to select the type of acceleration of
       the distance calculations.
    """
    if box is not None:
        box = np.asarray(box, dtype=np.float32)
        if box.shape[0] != 6:
            raise ValueError(
                "Box Argument is of incompatible type. The "
                "dimension should be either None or of the form "
                "[lx, ly, lz, alpha, beta, gamma]"
            )
    # The check_coords decorator made sure that reference is an
    # array of positions. Mypy does not know about that so we have to
    # tell it.
    reference_positions: npt.NDArray = reference  # type: ignore
    function = _determine_method_self(
        reference_positions,
        max_cutoff,
        min_cutoff=min_cutoff,
        box=box,
        method=method,
    )

    if function.__name__ == "_nsgrid_capped_self":
        return function(
            reference,
            max_cutoff,
            min_cutoff=min_cutoff,
            box=box,
            return_distances=return_distances,
        )
    else:
        return function(
            reference,
            max_cutoff,
            min_cutoff=min_cutoff,
            box=box,
            return_distances=return_distances,
            backend=backend,
        )


def _determine_method_self(
    reference: npt.NDArray,
    max_cutoff: float,
    min_cutoff: Optional[float] = None,
    box: Optional[npt.NDArray] = None,
    method: Optional[str] = None,
):
    """Guesses the fastest method for capped distance calculations based on the
    size of the `reference` coordinate set and the relative size of the target
    volume.

    Parameters
    ----------
    reference : numpy.ndarray
        Reference coordinate array with shape ``(3,)`` or ``(n, 3)``.
    max_cutoff : float
        Maximum cutoff distance between `reference` coordinates.
    min_cutoff : float, optional
        Minimum cutoff distance between `reference` coordinates.
    box : numpy.ndarray
        The unitcell dimensions of the system, which can be orthogonal or
        triclinic and must be provided in the same format as returned by
        :attr:`MDAnalysis.coordinates.timestep.Timestep.dimensions`:
        ``[lx, ly, lz, alpha, beta, gamma]``.
    method : {'bruteforce', 'nsgrid', 'pkdtree'}, optional
        Keyword to override the automatic guessing of the employed search
        method.

    Returns
    -------
    function : callable
        The function implementing the guessed (or deliberatly chosen) method.


    .. versionchanged:: 1.0.1
       nsgrid was temporarily removed and replaced with pkdtree due to issues
       relating to its reliability and accuracy (Issues #2919, #2229, #2345,
       #2670, #2930)
    .. versionchanged:: 1.0.2
       enabled nsgrid again
    """
    methods = {
        "bruteforce": _bruteforce_capped_self,
        "pkdtree": _pkdtree_capped_self,
        "nsgrid": _nsgrid_capped_self,
    }

    if method is not None:
        return methods[method.lower()]

    if len(reference) < 100:
        return methods["bruteforce"]

    if box is None:
        min_dim = np.array([reference.min(axis=0)])
        max_dim = np.array([reference.max(axis=0)])
        size = max_dim.max(axis=0) - min_dim.min(axis=0)
    elif np.all(box[3:] == 90.0):
        size = box[:3]
    else:
        tribox = triclinic_vectors(box)
        size = tribox.max(axis=0) - tribox.min(axis=0)

    if max_cutoff < 0.03 * size.min():
        return methods["pkdtree"]
    else:
        return methods["nsgrid"]


@check_coords(
    "reference",
    enforce_copy=False,
    reduce_result_if_single=False,
    allow_atomgroup=True,
)
def _bruteforce_capped_self(
    reference: Union[npt.NDArray, "AtomGroup"],
    max_cutoff: float,
    min_cutoff: Optional[float] = None,
    box: Optional[npt.NDArray] = None,
    return_distances: Optional[bool] = True,
    backend: Optional[str] = "serial",
):
    """Capped distance evaluations using a brute force method.

    Computes and returns an array containing pairs of indices corresponding to
    entries in the `reference` array which are separated by a distance lying
    within the specified cutoff(s). Employs naive distance computations (brute
    force) to find relevant distances. Optionally, these distances can be
    returned as well.

    If the optional argument `box` is supplied, the minimum image convention is
    applied when calculating distances. Either orthogonal or triclinic boxes are
    supported.

    Parameters
    ----------
    reference : numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Reference coordinate array with shape ``(3,)`` or ``(n, 3)`` (dtype will
        be converted to ``numpy.float32`` internally). Also accepts an
        :class:`~MDAnalysis.core.groups.AtomGroup`.
    max_cutoff : float
        Maximum cutoff distance between `reference` coordinates.
    min_cutoff : float, optional
        Minimum cutoff distance between `reference` coordinates.
    box : numpy.ndarray, optional
        The unitcell dimensions of the system, which can be orthogonal or
        triclinic and must be provided in the same format as returned by
        :attr:`MDAnalysis.coordinates.timestep.Timestep.dimensions`:
        ``[lx, ly, lz, alpha, beta, gamma]``.
    return_distances : bool, optional
        If set to ``True``, distances will also be returned.
    backend : {'serial', 'OpenMP', 'distopia'}, optional
        Keyword selecting the type of acceleration.

    Returns
    -------
    pairs : numpy.ndarray (``dtype=numpy.int64``, ``shape=(n_pairs, 2)``)
        Pairs of indices, corresponding to coordinates in the `reference` array
        such that the distance between them lies within the interval
        (`min_cutoff`, `max_cutoff`].
        Each row in `pairs` is an index pair ``[i, j]`` corresponding to the
        ``i``-th and the ``j``-th coordinate in `reference`.
    distances : numpy.ndarray (``dtype=numpy.float64``, ``shape=(n_pairs,)``), optional
        Distances corresponding to each pair of indices. Only returned if
        `return_distances` is ``True``. ``distances[k]`` corresponds to the
        ``k``-th pair returned in `pairs` and gives the distance between the
        coordinates ``reference[pairs[k, 0]]`` and
        ``configuration[pairs[k, 1]]``.

    .. versionchanged:: 0.20.0
       Added `return_distances` keyword.
    .. versionchanged:: 2.3.0
       Can now accept an :class:`~MDAnalysis.core.groups.AtomGroup` as an
       argument in any position and checks inputs using type hinting.
    .. versionchanged:: 2.10.0
       Added the "backend" argument to select the type of acceleration of
       the distance calculations.
    """
    # Default return values (will be overwritten only if pairs are found):
    pairs = np.empty((0, 2), dtype=np.intp)
    distances = np.empty((0,), dtype=np.float64)

    N = len(reference)
    # We're searching within a single coordinate set, so we need at least two
    # coordinates to find distances between them.
    if N > 1:
        distvec = self_distance_array(reference, box=box, backend=backend)
        dist = np.full((N, N), np.finfo(np.float64).max, dtype=np.float64)
        dist[np.triu_indices(N, 1)] = distvec

        if min_cutoff is not None:
            mask = np.where((dist <= max_cutoff) & (dist > min_cutoff))
        else:
            mask = np.where((dist <= max_cutoff))

        if mask[0].size > 0:
            pairs = np.c_[mask[0], mask[1]]
            distances = dist[mask]
    if return_distances:
        return pairs, distances
    return pairs


@check_coords(
    "reference",
    enforce_copy=False,
    reduce_result_if_single=False,
    allow_atomgroup=True,
)
def _pkdtree_capped_self(
    reference: Union[npt.NDArray, "AtomGroup"],
    max_cutoff: float,
    min_cutoff: Optional[float] = None,
    box: Optional[npt.NDArray] = None,
    return_distances: Optional[bool] = True,
    backend: Optional[str] = "serial",
):
    """Capped distance evaluations using a KDtree method.

    Computes and returns an array containing pairs of indices corresponding to
    entries in the `reference` array which are separated by a distance lying
    within the specified cutoff(s). Employs a (periodic) KDtree algorithm to
    find relevant distances. Optionally, these distances can be returned as
    well.

    If the optional argument `box` is supplied, the minimum image convention is
    applied when calculating distances. Either orthogonal or triclinic boxes are
    supported.

    Parameters
    ----------
    reference : numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Reference coordinate array with shape ``(3,)`` or ``(n, 3)`` (dtype will
        be converted to ``numpy.float32`` internally). Also accepts an
        :class:`~MDAnalysis.core.groups.AtomGroup`.
    max_cutoff : float
        Maximum cutoff distance between `reference` coordinates.
    min_cutoff : float, optional
        Minimum cutoff distance between `reference` coordinates.
    box : numpy.ndarray, optional
        The unitcell dimensions of the system, which can be orthogonal or
        triclinic and must be provided in the same format as returned by
        :attr:`MDAnalysis.coordinates.timestep.Timestep.dimensions`:
        ``[lx, ly, lz, alpha, beta, gamma]``.
    return_distances : bool, optional
        If set to ``True``, distances will also be returned.
    backend : {'serial', 'OpenMP', 'distopia'}, optional
        Keyword selecting the type of acceleration.

    Returns
    -------
    pairs : numpy.ndarray (``dtype=numpy.int64``, ``shape=(n_pairs, 2)``)
        Pairs of indices, corresponding to coordinates in the `reference` array
        such that the distance between them lies within the interval
        (`min_cutoff`, `max_cutoff`].
        Each row in `pairs` is an index pair ``[i, j]`` corresponding to the
        ``i``-th and the ``j``-th coordinate in `reference`.
    distances : numpy.ndarray (``dtype=numpy.float64``, ``shape=(n_pairs,)``)
        Distances corresponding to each pair of indices. Only returned if
        `return_distances` is ``True``. ``distances[k]`` corresponds to the
        ``k``-th pair returned in `pairs` and gives the distance between
        the coordinates ``reference[pairs[k, 0]]`` and
        ``reference[pairs[k, 1]]``.

    .. versionchanged:: 0.20.0
       Added `return_distances` keyword.
    .. versionchanged:: 2.3.0
       Can now accept an :class:`~MDAnalysis.core.groups.AtomGroup` as an
       argument in any position and checks inputs using type hinting.
    .. versionchanged:: 2.10.0
       Added the "backend" argument to select the type of acceleration of
       the distance calculations.
    """
    from .pkdtree import (
        PeriodicKDTree,
    )  # must be here to avoid circular import

    # Default return values (will be overwritten only if pairs are found):
    pairs = np.empty((0, 2), dtype=np.intp)
    distances = np.empty((0,), dtype=np.float64)

    # We're searching within a single coordinate set, so we need at least two
    # coordinates to find distances between them.
    if len(reference) > 1:
        kdtree = PeriodicKDTree(box=box)
        cut = max_cutoff if box is not None else None
        kdtree.set_coords(reference, cutoff=cut)
        _pairs = kdtree.search_pairs(max_cutoff)
        if _pairs.size > 0:
            pairs = _pairs
            if return_distances or (min_cutoff is not None):
                refA, refB = pairs[:, 0], pairs[:, 1]
                distances = calc_bonds(
                    reference[refA], reference[refB], box=box, backend=backend
                )
                if min_cutoff is not None:
                    idx = distances > min_cutoff
                    pairs, distances = pairs[idx], distances[idx]
    if return_distances:
        return pairs, distances
    return pairs


@check_coords(
    "reference",
    enforce_copy=False,
    reduce_result_if_single=False,
    allow_atomgroup=True,
)
def _nsgrid_capped_self(
    reference: Union[npt.NDArray, "AtomGroup"],
    max_cutoff: float,
    min_cutoff: Optional[float] = None,
    box: Optional[npt.NDArray] = None,
    return_distances: Optional[bool] = True,
):
    """Capped distance evaluations using a grid-based search method.

    Computes and returns an array containing pairs of indices corresponding to
    entries in the `reference` array which are separated by a distance lying
    within the specified cutoff(s). Employs a grid-based search algorithm to
    find relevant distances. Optionally, these distances can be returned as
    well.

    If the optional argument `box` is supplied, the minimum image convention is
    applied when calculating distances. Either orthogonal or triclinic boxes are
    supported.

    Parameters
    ----------
    reference : numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Reference coordinate array with shape ``(3,)`` or ``(n, 3)`` (dtype will
        be converted to ``numpy.float32`` internally).  Also accepts an
        :class:`~MDAnalysis.core.groups.AtomGroup`.
    max_cutoff : float
        Maximum cutoff distance between `reference` coordinates.
    min_cutoff : float, optional
        Minimum cutoff distance between `reference` coordinates.
    box : numpy.ndarray, optional
        The unitcell dimensions of the system, which can be orthogonal or
        triclinic and must be provided in the same format as returned by
        :attr:`MDAnalysis.coordinates.timestep.Timestep.dimensions`:
        ``[lx, ly, lz, alpha, beta, gamma]``.

    Returns
    -------
    pairs : numpy.ndarray (``dtype=numpy.int64``, ``shape=(n_pairs, 2)``)
        Pairs of indices, corresponding to coordinates in the `reference` array
        such that the distance between them lies within the interval
        (`min_cutoff`, `max_cutoff`].
        Each row in `pairs` is an index pair ``[i, j]`` corresponding to the
        ``i``-th and the ``j``-th coordinate in `reference`.
    distances : numpy.ndarray, optional
        Distances corresponding to each pair of indices. Only returned if
        `return_distances` is ``True``. ``distances[k]`` corresponds to the
        ``k``-th pair returned in `pairs` and gives the distance between the
        coordinates ``reference[pairs[k, 0]]`` and
        ``configuration[pairs[k, 1]]``.

    .. versionchanged:: 0.20.0
       Added `return_distances` keyword.
    .. versionchanged:: 2.3.0
       Can now accept an :class:`~MDAnalysis.core.groups.AtomGroup` as an
       argument in any position and checks inputs using type hinting.
    """
    # Default return values (will be overwritten only if pairs are found):
    pairs = np.empty((0, 2), dtype=np.intp)
    distances = np.empty((0,), dtype=np.float64)

    # We're searching within a single coordinate set, so we need at least two
    # coordinates to find distances between them.
    if len(reference) > 1:
        if box is None:
            # create a pseudobox
            # define the max range
            # and supply the pseudobox
            # along with only one set of coordinates
            pseudobox = np.zeros(6, dtype=np.float32)
            lmax = reference.max(axis=0)
            lmin = reference.min(axis=0)
            # Using maximum dimension as the box size
            boxsize = (lmax - lmin).max()
            # to avoid failures of very close particles
            # but with larger cutoff
            if boxsize < 2 * max_cutoff:
                # just enough box size so that NSGrid doesnot fails
                sizefactor = 2.2 * max_cutoff / boxsize
            else:
                sizefactor = 1.2
            pseudobox[:3] = sizefactor * boxsize
            pseudobox[3:] = 90.0
            shiftref = reference.copy()
            # Extra padding near the origin
            shiftref -= lmin - 0.1 * boxsize
            gridsearch = FastNS(max_cutoff, shiftref, box=pseudobox, pbc=False)
            results = gridsearch.self_search()
        else:
            gridsearch = FastNS(max_cutoff, reference, box=box)
            results = gridsearch.self_search()

        pairs = results.get_pairs()
        if return_distances or (min_cutoff is not None):
            distances = results.get_pair_distances()
            if min_cutoff is not None:
                idx = distances > min_cutoff
                pairs, distances = pairs[idx], distances[idx]

    if return_distances:
        return pairs, distances
    return pairs


@check_coords("coords")
def transform_RtoS(coords, box, backend="serial"):
    """Transform an array of coordinates from real space to S space (a.k.a.
    lambda space)

    S space represents fractional space within the unit cell for this system.

    Reciprocal operation to :meth:`transform_StoR`.

    Parameters
    ----------
    coords : numpy.ndarray
        A ``(3,)`` or ``(n, 3)`` array of coordinates (dtype is arbitrary, will
        be converted to ``numpy.float32`` internally).
    box : numpy.ndarray
        The unitcell dimensions of the system, which can be orthogonal or
        triclinic and must be provided in the same format as returned by
        :attr:`MDAnalysis.coordinates.timestep.Timestep.dimensions`:
        ``[lx, ly, lz, alpha, beta, gamma]``.
    backend : {'serial', 'OpenMP'}, optional
        Keyword selecting the type of acceleration.

    Returns
    -------
    newcoords : numpy.ndarray (``dtype=numpy.float32``, ``shape=coords.shape``)
        An array containing fractional coordiantes.


    .. versionchanged:: 0.13.0
       Added *backend* keyword.
    .. versionchanged:: 0.19.0
       Internal dtype conversion of input coordinates to ``numpy.float32``.
       Now also accepts (and, likewise, returns) a single coordinate.
    """
    if len(coords) == 0:
        return coords
    boxtype, box = check_box(box)
    if boxtype == "ortho":
        box = np.diag(box)
    box = box.astype(np.float64)

    # Create inverse matrix of box
    # need order C here
    inv = np.array(np.linalg.inv(box), order="C")

    _run("coord_transform", args=(coords, inv), backend=backend)

    return coords


@check_coords("coords")
def transform_StoR(coords, box, backend="serial"):
    """Transform an array of coordinates from S space into real space.

    S space represents fractional space within the unit cell for this system.

    Reciprocal operation to :meth:`transform_RtoS`

    Parameters
    ----------
    coords : numpy.ndarray
        A ``(3,)`` or ``(n, 3)`` array of coordinates (dtype is arbitrary, will
        be converted to ``numpy.float32`` internally).
    box : numpy.ndarray
        The unitcell dimensions of the system, which can be orthogonal or
        triclinic and must be provided in the same format as returned by
        :attr:`MDAnalysis.coordinates.timestep.Timestep.dimensions`:
        ``[lx, ly, lz, alpha, beta, gamma]``.
    backend : {'serial', 'OpenMP'}, optional
        Keyword selecting the type of acceleration.

    Returns
    -------
    newcoords : numpy.ndarray (``dtype=numpy.float32``, ``shape=coords.shape``)
        An array containing real space coordiantes.


    .. versionchanged:: 0.13.0
       Added *backend* keyword.
    .. versionchanged:: 0.19.0
       Internal dtype conversion of input coordinates to ``numpy.float32``.
       Now also accepts (and, likewise, returns) a single coordinate.
    """
    if len(coords) == 0:
        return coords
    boxtype, box = check_box(box)
    if boxtype == "ortho":
        box = np.diag(box)
    box = box.astype(np.float64)

    _run("coord_transform", args=(coords, box), backend=backend)
    return coords


@check_coords("coords1", "coords2", allow_atomgroup=True)
def calc_bonds(
    coords1: Union[npt.NDArray, "AtomGroup"],
    coords2: Union[npt.NDArray, "AtomGroup"],
    box: Optional[npt.NDArray] = None,
    result: Optional[npt.NDArray] = None,
    backend: str = "serial",
) -> npt.NDArray:
    """Calculates the bond lengths between pairs of atom positions from the two
    coordinate arrays `coords1` and `coords2`, which must contain the same
    number of coordinates. ``coords1[i]`` and ``coords2[i]`` represent the
    positions of atoms connected by the ``i``-th bond. If single coordinates are
    supplied, a single distance will be returned.

    In comparison to :meth:`distance_array` and :meth:`self_distance_array`,
    which calculate distances between all possible combinations of coordinates,
    :meth:`calc_bonds` only calculates distances between pairs of coordinates,
    similar to::

       numpy.linalg.norm(a - b) for a, b in zip(coords1, coords2)

    If the optional argument `box` is supplied, the minimum image convention is
    applied when calculating distances. Either orthogonal or triclinic boxes are
    supported.

    If a numpy array of dtype ``numpy.float64`` with shape ``(n,)`` (for ``n``
    coordinate pairs) is provided in `result`, then this preallocated array is
    filled. This can speed up calculations.

    Parameters
    ----------
    coords1 : numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Coordinate array of shape ``(3,)`` or ``(n, 3)`` for one half of a
        single or ``n`` bonds, respectively (dtype is arbitrary, will be
        converted to ``numpy.float32`` internally).  Also accepts an
        :class:`~MDAnalysis.core.groups.AtomGroup`.
    coords2 : numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Coordinate array of shape ``(3,)`` or ``(n, 3)`` for the other half of
        a single or ``n`` bonds, respectively (dtype is arbitrary, will be
        converted to ``numpy.float32`` internally). Also accepts an
        :class:`~MDAnalysis.core.groups.AtomGroup`.
    box : numpy.ndarray, optional
        The unitcell dimensions of the system, which can be orthogonal or
        triclinic and must be provided in the same format as returned by
        :attr:`MDAnalysis.coordinates.timestep.Timestep.dimensions`:
        ``[lx, ly, lz, alpha, beta, gamma]``.
    result : numpy.ndarray, optional
        Preallocated result array of dtype ``numpy.float64`` and shape ``(n,)``
        (for ``n`` coordinate pairs). Avoids recreating the array in repeated
        function calls.
    backend : {'serial', 'OpenMP', 'distopia'}, optional
        Keyword selecting the type of acceleration. Defaults to 'serial'.

    Returns
    -------
    bondlengths : numpy.ndarray (``dtype=numpy.float64``, ``shape=(n,)``) or
        numpy.float64 Array containing the bond lengths between each pair of
        coordinates. If two single coordinates were supplied, their distance is
        returned as a single number instead of an array.


    .. versionadded:: 0.8
    .. versionchanged:: 0.13.0
       Added *backend* keyword.
    .. versionchanged:: 0.19.0
       Internal dtype conversion of input coordinates to ``numpy.float32``.
       Now also accepts single coordinates as input.
    .. versionchanged:: 2.3.0
       Can now accept an :class:`~MDAnalysis.core.groups.AtomGroup` as an
       argument in any position and checks inputs using type hinting.
    .. versionchanged:: 2.5.0
       Can now optionally use the fast distance functions from distopia
    """
    numatom = coords1.shape[0]
    bondlengths = _check_result_array(result, (numatom,))
    if backend == "distopia":
        # distopia requires that all the input arrays are the same type,
        # while MDAnalysis allows for mixed types, this should be changed
        # pre 3.0.0 release see issue #3707
        bondlengths = bondlengths.astype(np.float32)
        box = np.asarray(box).astype(np.float32) if box is not None else None

    if numatom > 0:
        if box is not None:
            boxtype, box = check_box(box)
            if boxtype == "ortho":
                _run(
                    "calc_bond_distance_ortho",
                    args=(coords1, coords2, box, bondlengths),
                    backend=backend,
                )
            else:
                _run(
                    "calc_bond_distance_triclinic",
                    args=(coords1, coords2, box, bondlengths),
                    backend=backend,
                )
        else:
            _run(
                "calc_bond_distance",
                args=(coords1, coords2, bondlengths),
                backend=backend,
            )
    if backend == "distopia":
        # mda expects the result to be in float64, so we need to convert it back
        # to float64, change for 3.0, see #3707
        bondlengths = bondlengths.astype(np.float64)
        if result is not None:
            result[:] = bondlengths

    return bondlengths


@check_coords("coords1", "coords2", "coords3", allow_atomgroup=True)
def calc_angles(
    coords1: Union[npt.NDArray, "AtomGroup"],
    coords2: Union[npt.NDArray, "AtomGroup"],
    coords3: Union[npt.NDArray, "AtomGroup"],
    box: Optional[npt.NDArray] = None,
    result: Optional[npt.NDArray] = None,
    backend: str = "serial",
) -> npt.NDArray:
    """Calculates the angles formed between triplets of atom positions from the
    three coordinate arrays `coords1`, `coords2`, and `coords3`. All coordinate
    arrays must contain the same number of coordinates.

    The coordinates in `coords2` represent the apices of the angles::

            2---3
           /
          1

    Configurations where the angle is undefined (e.g., when coordinates 1 or 3
    of a triplet coincide with coordinate 2) result in a value of **zero** for
    that angle.

    If the optional argument `box` is supplied, periodic boundaries are taken
    into account when constructing the connecting vectors between coordinates,
    i.e., the minimum image convention is applied for the vectors forming the
    angles. Either orthogonal or triclinic boxes are supported.

    If a numpy array of dtype ``numpy.float64`` with shape ``(n,)`` (for ``n``
    coordinate triplets) is provided in `result`, then this preallocated array
    is filled. This can speed up calculations.

    Parameters
    ----------
    coords1 : numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Array of shape ``(3,)`` or ``(n, 3)`` containing the coordinates of one
        side of a single or ``n`` angles, respectively (dtype is arbitrary, will
        be converted to ``numpy.float32`` internally). Also accepts an
        :class:`~MDAnalysis.core.groups.AtomGroup`.
    coords2 :  numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Array of shape ``(3,)`` or ``(n, 3)`` containing the coordinates of the
        apices of a single or ``n`` angles, respectively (dtype is arbitrary,
        will be converted to ``numpy.float32`` internally). Also accepts an
        :class:`~MDAnalysis.core.groups.AtomGroup`.
    coords3 :  numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Array of shape ``(3,)`` or ``(n, 3)`` containing the coordinates of the
        other side of a single or ``n`` angles, respectively (dtype is
        arbitrary, will be converted to ``numpy.float32`` internally).
        Also accepts an :class:`~MDAnalysis.core.groups.AtomGroup`.
    box : numpy.ndarray, optional
        The unitcell dimensions of the system, which can be orthogonal or
        triclinic and must be provided in the same format as returned by
        :attr:`MDAnalysis.coordinates.timestep.Timestep.dimensions`:
        ``[lx, ly, lz, alpha, beta, gamma]``.
    result : numpy.ndarray, optional
        Preallocated result array of dtype ``numpy.float64`` and shape ``(n,)``
        (for ``n`` coordinate triplets). Avoids recreating the array in repeated
        function calls.
    backend : {'serial', 'OpenMP', 'distopia'}, optional
        Keyword selecting the type of acceleration.

    Returns
    -------
    angles : numpy.ndarray (``dtype=numpy.float64``, ``shape=(n,)``) or numpy.float64
        Array containing the angles between each triplet of coordinates. Values
        are returned in radians (rad). If three single coordinates were
        supplied, the angle is returned as a single number instead of an array.


    .. versionadded:: 0.8
    .. versionchanged:: 0.9.0
       Added optional box argument to account for periodic boundaries in
       calculation
    .. versionchanged:: 0.13.0
       Added *backend* keyword.
    .. versionchanged:: 0.19.0
       Internal dtype conversion of input coordinates to ``numpy.float32``.
       Now also accepts single coordinates as input.
    .. versionchanged:: 2.3.0
       Can now accept an :class:`~MDAnalysis.core.groups.AtomGroup` as an
       argument in any position and checks inputs using type hinting.
    """
    numatom = coords1.shape[0]
    angles = _check_result_array(result, (numatom,))

    if backend == "distopia":
        # distopia requires that all the input arrays are the same type,
        # while MDAnalysis allows for mixed types, this should be changed
        # pre 3.0.0 release see issue #3707
        angles = angles.astype(np.float32)
        box = np.asarray(box).astype(np.float32) if box is not None else None

    if numatom > 0:
        if box is not None:
            boxtype, box = check_box(box)
            if boxtype == "ortho":
                _run(
                    "calc_angle_ortho",
                    args=(coords1, coords2, coords3, box, angles),
                    backend=backend,
                )
            else:
                _run(
                    "calc_angle_triclinic",
                    args=(coords1, coords2, coords3, box, angles),
                    backend=backend,
                )
        else:
            _run(
                "calc_angle",
                args=(coords1, coords2, coords3, angles),
                backend=backend,
            )

    if backend == "distopia":
        # mda expects the result to be in float64, so we need to convert it back
        # to float64, change for 3.0, see #3707
        angles = angles.astype(np.float64)
        if result is not None:
            result[:] = angles
    return angles


@check_coords("coords1", "coords2", "coords3", "coords4", allow_atomgroup=True)
def calc_dihedrals(
    coords1: Union[npt.NDArray, "AtomGroup"],
    coords2: Union[npt.NDArray, "AtomGroup"],
    coords3: Union[npt.NDArray, "AtomGroup"],
    coords4: Union[npt.NDArray, "AtomGroup"],
    box: Optional[npt.NDArray] = None,
    result: Optional[npt.NDArray] = None,
    backend: str = "serial",
) -> npt.NDArray:
    r"""Calculates the dihedral angles formed between quadruplets of positions
    from the four coordinate arrays `coords1`, `coords2`, `coords3`, and
    `coords4`, which must contain the same number of coordinates.

    The dihedral angle formed by a quadruplet of positions (1,2,3,4) is
    calculated around the axis connecting positions 2 and 3 (i.e., the angle
    between the planes spanned by positions (1,2,3) and (2,3,4))::

                  4
                  |
            2-----3
           /
          1

    If all coordinates lie in the same plane, the cis configuration corresponds
    to a dihedral angle of zero, and the trans configuration to :math:`\pi`
    radians (180 degrees). Configurations where the dihedral angle is undefined
    (e.g., when all coordinates lie on the same straight line) result in a value
    of ``nan`` (not a number) for that dihedral.

    If the optional argument `box` is supplied, periodic boundaries are taken
    into account when constructing the connecting vectors between coordinates,
    i.e., the minimum image convention is applied for the vectors forming the
    dihedral angles. Either orthogonal or triclinic boxes are supported.

    If a numpy array of dtype ``numpy.float64`` with shape ``(n,)`` (for ``n``
    coordinate quadruplets) is provided in `result` then this preallocated array
    is filled. This can speed up calculations.

    Parameters
    ----------
    coords1 : numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Coordinate array of shape ``(3,)`` or ``(n, 3)`` containing the 1st
        positions in dihedrals (dtype is arbitrary, will be converted to
        ``numpy.float32`` internally).  Also accepts an
        :class:`~MDAnalysis.core.groups.AtomGroup`.
    coords2 : numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Coordinate array of shape ``(3,)`` or ``(n, 3)`` containing the 2nd
        positions in dihedrals (dtype is arbitrary, will be converted to
        ``numpy.float32`` internally).  Also accepts an
        :class:`~MDAnalysis.core.groups.AtomGroup`.
    coords3 : numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Coordinate array of shape ``(3,)`` or ``(n, 3)`` containing the 3rd
        positions in dihedrals (dtype is arbitrary, will be converted to
        ``numpy.float32`` internally).  Also accepts an
        :class:`~MDAnalysis.core.groups.AtomGroup`.
    coords4 : numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Coordinate array of shape ``(3,)`` or ``(n, 3)`` containing the 4th
        positions in dihedrals (dtype is arbitrary, will be converted to
        ``numpy.float32`` internally).  Also accepts an
        :class:`~MDAnalysis.core.groups.AtomGroup`.
    box : numpy.ndarray, optional
        The unitcell dimensions of the system, which can be orthogonal or
        triclinic and must be provided in the same format as returned by
        :attr:`MDAnalysis.coordinates.timestep.Timestep.dimensions`:
        ``[lx, ly, lz, alpha, beta, gamma]``.
    result : numpy.ndarray, optional
        Preallocated result array of dtype ``numpy.float64`` and shape ``(n,)``
        (for ``n`` coordinate quadruplets). Avoids recreating the array in
        repeated function calls.
    backend : {'serial', 'OpenMP', 'distopia'}, optional
        Keyword selecting the type of acceleration.

    Returns
    -------
    dihedrals : numpy.ndarray (``dtype=numpy.float64``, ``shape=(n,)``) or numpy.float64
        Array containing the dihedral angles formed by each quadruplet of
        coordinates. Values are returned in radians (rad). If four single
        coordinates were supplied, the dihedral angle is returned as a single
        number instead of an array. The range of dihedral angle is
        :math:`(-\pi, \pi)`.


    .. versionadded:: 0.8
    .. versionchanged:: 0.9.0
       Added optional box argument to account for periodic boundaries in
       calculation
    .. versionchanged:: 0.11.0
       Renamed from calc_torsions to calc_dihedrals
    .. versionchanged:: 0.13.0
       Added *backend* keyword.
    .. versionchanged:: 0.19.0
       Internal dtype conversion of input coordinates to ``numpy.float32``.
       Now also accepts single coordinates as input.
    .. versionchanged:: 2.3.0
       Can now accept an :class:`~MDAnalysis.core.groups.AtomGroup` as an
       argument in any position and checks inputs using type hinting.
    """
    numatom = coords1.shape[0]
    dihedrals = _check_result_array(result, (numatom,))

    if backend == "distopia":
        # distopia requires that all the input arrays are the same type,
        # while MDAnalysis allows for mixed types, this should be changed
        # pre 3.0.0 release see issue #3707
        dihedrals = dihedrals.astype(np.float32)
        box = np.asarray(box).astype(np.float32) if box is not None else None

    if numatom > 0:
        if box is not None:
            boxtype, box = check_box(box)
            if boxtype == "ortho":
                _run(
                    "calc_dihedral_ortho",
                    args=(coords1, coords2, coords3, coords4, box, dihedrals),
                    backend=backend,
                )
            else:
                _run(
                    "calc_dihedral_triclinic",
                    args=(coords1, coords2, coords3, coords4, box, dihedrals),
                    backend=backend,
                )
        else:
            _run(
                "calc_dihedral",
                args=(coords1, coords2, coords3, coords4, dihedrals),
                backend=backend,
            )
    if backend == "distopia":
        # mda expects the result to be in float64, so we need to convert it back
        # to float64, change for 3.0, see #3707
        dihedrals = dihedrals.astype(np.float64)
        if result is not None:
            result[:] = dihedrals
    return dihedrals


@check_coords("coords", allow_atomgroup=True)
def apply_PBC(
    coords: Union[npt.NDArray, "AtomGroup"],
    box: Optional[npt.NDArray] = None,
    backend: str = "serial",
) -> npt.NDArray:
    """Moves coordinates into the primary unit cell.

    Parameters
    ----------
    coords : numpy.ndarray or :class:`~MDAnalysis.core.groups.AtomGroup`
        Coordinate array of shape ``(3,)`` or ``(n, 3)`` (dtype is arbitrary,
        will be converted to ``numpy.float32`` internally). Also accepts an
        :class:`~MDAnalysis.core.groups.AtomGroup`.
    box : numpy.ndarray
        The unitcell dimensions of the system, which can be orthogonal or
        triclinic and must be provided in the same format as returned by
        :attr:`MDAnalysis.coordinates.timestep.Timestep.dimensions`:
        ``[lx, ly, lz, alpha, beta, gamma]``.
    backend : {'serial', 'OpenMP'}, optional
        Keyword selecting the type of acceleration.

    Returns
    -------
    newcoords : numpy.ndarray  (``dtype=numpy.float32``, ``shape=coords.shape``)
        Array containing coordinates that all lie within the primary unit cell
        as defined by `box`.


    .. versionadded:: 0.8
    .. versionchanged:: 0.13.0
       Added *backend* keyword.
    .. versionchanged:: 0.19.0
       Internal dtype conversion of input coordinates to ``numpy.float32``.
       Now also accepts (and, likewise, returns) single coordinates.
    .. versionchanged:: 2.3.0
       Can now accept an :class:`~MDAnalysis.core.groups.AtomGroup` as an
       argument in any position and checks inputs using type hinting.
    """
    # coords is an array, the check_coords decorator made sure of that.
    # Mypy, however, is not aware of that so we have to tell it explicitly.
    coords_array: npt.NDArray = coords  # type: ignore

    if len(coords_array) == 0:
        return coords_array
    boxtype, box = check_box(box)
    if boxtype == "ortho":
        _run("ortho_pbc", args=(coords_array, box), backend=backend)
    else:
        _run("triclinic_pbc", args=(coords_array, box), backend=backend)

    return coords_array


@check_coords("vectors", enforce_copy=False, enforce_dtype=False)
def minimize_vectors(vectors: npt.NDArray, box: npt.NDArray) -> npt.NDArray:
    """Apply minimum image convention to an array of vectors

    This function is required for calculating the correct vectors between two
    points.  A naive approach of ``ag1.positions - ag2.positions`` will not
    provide the minimum vectors between particles, even if all particles are
    within the primary unit cell (box).

    Parameters
    ----------
    vectors : numpy.ndarray
        Vector array of shape ``(n, 3)``, either float32 or float64.  These
        represent many vectors (such as between two particles).
    box : numpy.ndarray
        The unitcell dimensions of the system, which can be orthogonal or
        triclinic and must be provided in the same format as returned by
        :attr:`MDAnalysis.coordinates.timestep.Timestep.dimensions`:
        ``[lx, ly, lz, alpha, beta, gamma]``.

    Returns
    -------
    minimized_vectors : numpy.ndarray
        Same shape and dtype as input.  The vectors from the input, but
        minimized according to the size of the box.


    .. versionadded:: 2.1.0
    """
    boxtype, box = check_box(box)
    output = np.empty_like(vectors)

    # use box which is same precision as input vectors
    box = box.astype(vectors.dtype)

    if boxtype == "ortho":
        _minimize_vectors_ortho(vectors, box, output)
    else:
        _minimize_vectors_triclinic(vectors, box.ravel(), output)

    return output
